import depthai as dai
import pytest

from depthai_nodes import Keypoint, Keypoints


@pytest.fixture
def keypoint():
    return Keypoint()


@pytest.fixture
def keypoints():
    return Keypoints()


def test_keypoint_initialization(keypoint: Keypoint):
    assert keypoint.x is None
    assert keypoint.y is None
    assert keypoint.z == 0.0
    assert keypoint.confidence == -1.0


def test_keypoint_set_x(keypoint: Keypoint):
    keypoint.x = 0.5
    assert keypoint.x == 0.5

    with pytest.raises(TypeError):
        keypoint.x = "not a float"

    with pytest.raises(ValueError):
        keypoint.x = 1.5


def test_keypoint_set_y(keypoint: Keypoint):
    keypoint.y = 0.5
    assert keypoint.y == 0.5

    with pytest.raises(TypeError):
        keypoint.y = "not a float"

    with pytest.raises(ValueError):
        keypoint.y = 1.5


def test_keypoint_set_z(keypoint: Keypoint):
    keypoint.z = 0.5
    assert keypoint.z == 0.5

    with pytest.raises(TypeError):
        keypoint.z = "not a float"


def test_keypoint_set_confidence(keypoint: Keypoint):
    keypoint.confidence = 0.9
    assert keypoint.confidence == 0.9

    with pytest.raises(TypeError):
        keypoint.confidence = "not a float"

    with pytest.raises(ValueError):
        keypoint.confidence = 1.5


def test_keypoints_initialization(keypoints: Keypoints):
    assert keypoints.keypoints == []
    assert keypoints.transformation is None


def test_keypoints_set_keypoints(keypoints: Keypoints):
    kp1 = Keypoint()
    kp2 = Keypoint()
    keypoints_list = [kp1, kp2]
    keypoints.keypoints = keypoints_list
    assert keypoints.keypoints == keypoints_list

    with pytest.raises(TypeError):
        keypoints.keypoints = "not a list"

    with pytest.raises(ValueError):
        keypoints.keypoints = [kp1, "not a Keypoint"]


def test_keypoints_set_transformation(keypoints: Keypoints):
    transformation = dai.ImgTransformation()
    keypoints.transformation = transformation
    assert keypoints.transformation == transformation

    with pytest.raises(TypeError):
        keypoints.transformation = "not a dai.ImgTransformation"


def test_keypoints_set_transformation_none(keypoints: Keypoints):
    keypoints.transformation = None
    assert keypoints.transformation is None



import numpy as np


def unnormalize_image(image, normalize=True):
    """Un-normalize an image tensor by scaling it to the [0, 255] range.

    @param image: The normalized image tensor of shape (H, W, C) or (C, H, W).
    @type image: np.ndarray
    @param normalize: Whether to normalize the image tensor. Defaults to True.
    @type normalize: bool
    @return: The un-normalized image.
    @rtype: np.ndarray
    """
    # Normalize the image tensor to the range [0, 1]
    if normalize:
        min_val = image.min()
        max_val = image.max()
        if max_val != min_val:
            image = (image - min_val) / (max_val - min_val)
        else:
            # Handle the case where all values in the image are the same
            image = image - min_val

    # Scale to [0, 255] and clip the values to be in the proper range
    image = image * 255.0
    image = np.clip(image, 0, 255)

    # Convert to uint8
    image = image.astype(np.uint8)

    return image



import time
from enum import Enum
from typing import List, Optional, Tuple

import numpy as np

from depthai_nodes.logging import get_logger
from depthai_nodes.node.parsers.utils import sigmoid, xywh_to_xyxy
from depthai_nodes.node.parsers.utils.nms import nms

logger = get_logger(__name__)


class YOLOSubtype(str, Enum):
    V3 = "yolov3"
    V3T = "yolov3-tiny"
    V3U = "yolov3-u"
    V3UT = "yolov3-tinyu"
    V4 = "yolov4"
    V4T = "yolov4-tiny"
    V5 = "yolov5"
    V5U = "yolov5-u"
    V6 = "yolov6"
    V6R1 = "yolov6r1"
    V6R2 = "yolov6r2"  # NOTE: used by DAI but internally same as V6
    V7 = "yolov7"
    V8 = "yolov8"
    V9 = "yolov9"
    V10 = "yolov10"
    P = "yolo-p"
    GOLD = "yolo-gold"
    DEFAULT = ""


def make_grid_numpy(ny: int, nx: int, na: int) -> np.ndarray:
    """Create a grid of shape (1, na, ny, nx, 2)

    @param ny: Number of y coordinates.
    @type ny: int
    @param nx: Number of x coordinates.
    @type nx: int
    @param na: Number of anchors.
    @type na: int
    @return: Grid.
    @rtype: np.ndarray
    """
    yv, xv = np.meshgrid(np.arange(ny), np.arange(nx), indexing="ij")
    return np.stack((xv, yv), 2).reshape(1, na, ny, nx, 2)


def non_max_suppression(
    prediction: np.ndarray,
    conf_thres: float = 0.5,
    iou_thres: float = 0.45,
    classes: Optional[List] = None,
    num_classes: int = 1,
    agnostic: bool = False,
    multi_label: bool = False,
    max_det: int = 300,
    max_time_img: float = 0.05,
    max_nms: int = 30000,
    max_wh: int = 7680,
    kpts_mode: bool = False,
    det_mode: bool = False,
) -> List[np.ndarray]:
    """Performs Non-Maximum Suppression (NMS) on inference results.

    @param prediction: Prediction from the model, shape = (batch_size, boxes, xy+wh+...)
    @type prediction: np.ndarray
    @param conf_thres: Confidence threshold.
    @type conf_thres: float
    @param iou_thres: Intersection over union threshold.
    @type iou_thres: float
    @param classes: For filtering by classes.
    @type classes: Optional[List]
    @param num_classes: Number of classes.
    @type num_classes: int
    @param agnostic: Runs NMS on all boxes together rather than per class if True.
    @type agnostic: bool
    @param multi_label: Multilabel classification.
    @type multi_label: bool
    @param max_det: Limiting detections.
    @type max_det: int
    @param max_time_img: Maximum time for processing an image.
    @type max_time_img: float
    @param max_nms: Maximum number of boxes.
    @type max_nms: int
    @param max_wh: Maximum width and height.
    @type max_wh: int
    @param kpts_mode: Keypoints mode.
    @type kpts_mode: bool
    @param det_mode: Detection only mode. If True, the output will only contain bbox detections.
    @type det_mode: bool
    @return: An array of detections. If det_mode is False, the detections may include kpts or segmentation outputs.
    @rtype: List[np.ndarray]
    """
    bs = prediction.shape[0]  # batch size

    offset = prediction.shape[2] - num_classes
    num_classes_check = prediction.shape[2] - offset

    nm = prediction.shape[2] - num_classes - 5
    pred_candidates = prediction[..., 4] > conf_thres  # candidates

    # Check the parameters.
    assert (
        num_classes == num_classes_check
    ), f"Number of classes {num_classes} does not match the model {num_classes_check}"
    assert (
        0 <= conf_thres <= 1
    ), f"Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0"
    assert (
        0 <= iou_thres <= 1
    ), f"Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0"

    # Function settings.
    time_limit = 2.0 + max_time_img * bs  # seconds to quit after
    multi_label &= num_classes > 1  # multiple labels per box

    tik = time.time()
    output = [np.zeros((0, 6 + nm))] * prediction.shape[0]

    for img_idx, x in enumerate(prediction):  # image index, image inference
        x = x[pred_candidates[img_idx]]  # confidence

        # If no box remains, skip the next process.
        if not x.shape[0]:
            continue

        # (center x, center y, width, height) to (x1, y1, x2, y2)
        box = xywh_to_xyxy(x[:, :4])
        cls = x[:, 5 : 5 + num_classes]
        other = x[:, 5 + num_classes :]  # Either kpts or pos

        if multi_label:
            box_idx, class_idx = (cls > conf_thres).nonzero(as_tuple=False).T
            x = np.concatenate(
                (
                    box[box_idx],
                    x[box_idx, class_idx + 5, None],
                    class_idx[:, None],
                    other[box_idx, :],
                ),
                1,
            )
        else:  # Only keep the class with highest scores.
            class_idx = np.expand_dims(cls.argmax(1), 1)
            conf = cls.max(1, keepdims=True)
            x = np.concatenate((box, conf, class_idx, other), 1)[
                conf.flatten() > conf_thres
            ]

        # Filter by class, only keep boxes whose category is in classes.
        if classes is not None:
            x = x[(x[:, 5:6] == np.array(classes)).any(1)]

        # Check shape
        num_box = x.shape[0]  # number of boxes
        if not num_box:  # no boxes kept.
            continue
        elif num_box > max_nms:  # excess max boxes' number.
            x = x[x[:, 4].argsort()[:max_nms]]  # sort by confidence

        # Batched NMS
        class_offset = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = (
            x[:, :4] + class_offset,
            x[:, 4][..., np.newaxis],
        )  # boxes (offset by class), scores
        keep_box_idx = np.array(
            nms(np.hstack((boxes, scores)).astype(np.float32, copy=False), iou_thres)
        )

        if keep_box_idx.shape[0] > max_det:  # limit detections
            keep_box_idx = keep_box_idx[:max_det]

        output[img_idx] = x[keep_box_idx]
        if (time.time() - tik) > time_limit:
            logger.info(f"NMS cost time exceed the limited {time_limit}s.")
            break  # time limit exceeded

    return output


def parse_yolo_outputs(
    outputs: List[np.ndarray],
    strides: List[int],
    num_outputs: int,
    anchors: Optional[np.ndarray] = None,
    kpts: Optional[List[np.ndarray]] = None,
    det_mode: bool = False,
    subtype: YOLOSubtype = YOLOSubtype.DEFAULT,
) -> np.ndarray:
    """Parse all outputs of an YOLO model (all channels).

    @param outputs: List of outputs of an YOLO model.
    @type outputs: List[np.ndarray]
    @param strides: List of strides.
    @type strides: List[int]
    @param num_outputs: Number of outputs of the model.
    @type num_outputs: int
    @param anchors: An optional array of anchors.
    @type anchors: Optional[np.ndarray]
    @param kpts: An optional list of keypoints for each output.
    @type kpts: Optional[List[np.ndarray]]
    @param det_mode: Detection only mode.
    @type det_mode: bool
    @param subtype: YOLO version.
    @type subtype: YOLOSubtype
    @return: Parsed output.
    @rtype: np.ndarray
    """
    output = None

    for i, (out_head, stride) in enumerate(zip(outputs, strides)):
        kpt = kpts[i] if kpts else None
        anchors_head = anchors[i] if anchors is not None else None
        out = parse_yolo_output(
            out_head,
            stride,
            num_outputs,
            anchors_head,
            head_id=i,
            kpts=kpt,
            det_mode=det_mode,
            subtype=subtype,
        )
        output = out if output is None else np.concatenate((output, out), axis=1)

    return output


def parse_yolo_output(
    out: np.ndarray,
    stride: int,
    num_outputs: int,
    anchors: Optional[np.ndarray] = None,
    head_id: int = -1,
    kpts: Optional[np.ndarray] = None,
    det_mode: bool = False,
    subtype: YOLOSubtype = YOLOSubtype.DEFAULT,
) -> np.ndarray:
    """Parse a single channel output of an YOLO model.

    @param out: A single output of an YOLO model for the given channel.
    @type out: np.ndarray
    @param stride: Stride.
    @type stride: int
    @param num_outputs: Number of outputs of the model.
    @type num_outputs: int
    @param anchors: Anchors for the given head.
    @type anchors: Optional[np.ndarray]
    @param head_id: Head ID.
    @type head_id: int
    @param kpts: A single output of keypoints for the given channel.
    @type kpts: Optional[np.ndarray]
    @param det_mode: Detection only mode.
    @type det_mode: bool
    @param subtype: YOLO version.
    @type subtype: YOLOSubtype
    @return: Parsed output.
    @rtype: np.ndarray
    """
    na = (
        anchors.shape[0] // 2 if anchors is not None else 1
    )  # number of anchors per head
    bs, _, ny, nx = out.shape  # bs - batch size, ny|nx - y and x of grid cells

    if subtype in [
        YOLOSubtype.P,
        YOLOSubtype.V5,
        YOLOSubtype.V5U,
        YOLOSubtype.V7,
        YOLOSubtype.V3,
        YOLOSubtype.V3T,
        YOLOSubtype.V4,
        YOLOSubtype.V4T,
    ]:
        grid = make_grid_numpy(ny, nx, 1)
    else:
        grid = make_grid_numpy(ny, nx, na)

    out = out.reshape(bs, na, num_outputs, ny, nx).transpose((0, 1, 3, 4, 2))

    if anchors is not None:
        if isinstance(anchors, np.ndarray):
            anchors = anchors.reshape(bs, -1, 1, 1, 2)
        else:
            anchors = np.array(anchors).reshape(bs, -1, 1, 1, 2)
        assert (
            anchors.shape[1] == na
        ), f"Anchor shape mismatch at dimension 1: {anchors.shape[1]} vs {na}"

        if subtype in [YOLOSubtype.V3, YOLOSubtype.V3T]:
            c_xy = out[..., 0:2] + grid
            wh = np.exp(out[..., 2:4])
        elif subtype in [YOLOSubtype.V4, YOLOSubtype.V4T]:
            raise NotImplementedError("YOLOv4 is not supported yet")
        else:
            c_xy = out[..., 0:2] * 2 - 0.5 + grid
            wh = (out[..., 2:4] * 2) ** 2

        out[..., 0:2] = c_xy * stride
        out[..., 2:4] = wh * anchors
    else:
        if subtype == YOLOSubtype.V6R1:
            c_xy = out[..., 0:2] + grid
            wh = np.exp(out[..., 2:4])
        else:
            x1y1 = grid - out[..., 0:2] + 0.5
            x2y2 = grid + out[..., 2:4] + 0.5
            c_xy = (x1y1 + x2y2) / 2
            wh = x2y2 - x1y1
        out[..., 0:2] = c_xy * stride
        out[..., 2:4] = wh * stride

    if det_mode:
        # Detection
        out = out.reshape(bs, -1, num_outputs)
    elif kpts is None:
        # Segmentation
        x_coors = np.tile(np.arange(0, nx), (ny, 1))
        x_coors = np.repeat(x_coors[np.newaxis, np.newaxis, ..., np.newaxis], 1, axis=1)

        y_coors = np.tile(np.arange(0, ny)[np.newaxis, ...].T, (1, nx))
        y_coors = np.repeat(y_coors[np.newaxis, np.newaxis, ..., np.newaxis], 1, axis=1)

        ai = (
            np.ones((bs, na, ny, nx))
            * np.arange(na)[np.newaxis, ..., np.newaxis, np.newaxis]
        )
        ai = ai[..., np.newaxis]
        hi = np.ones((bs, na, ny, nx, 1)) * head_id

        out = np.concatenate((out, hi, ai, x_coors, y_coors), axis=4).reshape(
            bs, na * ny * nx, -1
        )
    else:
        # Keypoints

        # NOTE: For now we omit "guessing" if sigmoid is applied, should be better handled with flags in NNarchive/Parser
        # sigmoid_applied = np.all((kpts[:, 2::3] >= 0) & (kpts[:, 2::3] <= 1))
        # if not sigmoid_applied:
        #     kpts[:, 2::3] = sigmoid(kpts[:, 2::3])

        kpts[:, 2::3] = sigmoid(kpts[:, 2::3])
        kpts_out = kpts.transpose(0, 2, 1)
        out = out.reshape(bs, ny * nx, -1)
        out = np.concatenate((out, kpts_out), axis=2)

    return out


def parse_kpts(
    kpts: np.ndarray, n_keypoints: int, img_shape: Tuple[int, int]
) -> List[Tuple[int, int, float]]:
    """Parse keypoints.

    @param kpts: Result keypoints.
    @type kpts: np.ndarray
    @param n_keypoints: Number of keypoints.
    @type n_keypoints: int
    @param img_shape: Image shape of the model input in (height, width) format.
    @type img_shape: Tuple[int, int]
    @return: Parsed keypoints.
    @rtype: List[Tuple[int, int, float]]
    """
    h, w = img_shape
    kps = []
    ndim = len(kpts) // n_keypoints
    for idx in range(0, kpts.shape[0], ndim):
        x, y = kpts[idx] / w, kpts[idx + 1] / h
        conf = kpts[idx + 2] if ndim == 3 else 1.0
        kps.append((x, y, conf))
    return kps


def decode_yolo_output(
    yolo_outputs: List[np.ndarray],
    strides: List[int],
    anchors: Optional[np.ndarray] = None,
    kpts: Optional[List[np.ndarray]] = None,
    conf_thres: float = 0.5,
    iou_thres: float = 0.45,
    num_classes: int = 1,
    det_mode: bool = False,
    subtype: YOLOSubtype = YOLOSubtype.DEFAULT,
) -> np.ndarray:
    """Decode the output of an YOLO instance segmentation or pose estimation model.

    @param yolo_outputs: List of YOLO outputs.
    @type yolo_outputs: List[np.ndarray]
    @param strides: List of strides.
    @type strides: List[int]
    @param anchors: An optional array of anchors.
    @type anchors: Optional[np.ndarray]
    @param kpts: An optional list of keypoints.
    @type kpts: Optional[List[np.ndarray]]
    @param conf_thres: Confidence threshold.
    @type conf_thres: float
    @param iou_thres: Intersection over union threshold.
    @type iou_thres: float
    @param num_classes: Number of classes.
    @type num_classes: int
    @param det_mode: Detection only mode. If True, the output will only contain bbox
        detections.
    @type det_mode: bool
    @param subtype: YOLO version.
    @type subtype: YOLOSubtype
    @return: NMS output.
    @rtype: np.ndarray
    """
    num_outputs = num_classes + 5
    output = parse_yolo_outputs(
        yolo_outputs, strides, num_outputs, anchors, kpts, det_mode, subtype
    )
    output_nms = non_max_suppression(
        output,
        conf_thres=conf_thres,
        iou_thres=iou_thres,
        num_classes=num_classes,
        kpts_mode=kpts is not None,
        det_mode=det_mode,
    )[0]

    return output_nms



from typing import Any, Dict, List, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_keypoints_message
from depthai_nodes.node.parsers.keypoints import KeypointParser


class HRNetParser(KeypointParser):
    """Parser class for parsing the output of the HRNet pose estimation model. The code is inspired by https://github.com/ibaiGorordo/ONNX-HRNET-Human-Pose-Estimation.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    score_threshold : float
        Confidence score threshold for detected keypoints.
    label_names: Optional[List[str]]
        Label names for the keypoints.
    edges: Optional[List[Tuple[int, int]]]
        Keypoint connection pairs for visualizing the skeleton. Example:
            [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint
            1, keypoint 1 is connected to keypoint 2, etc.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing detected body keypoints.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        score_threshold: float = 0.5,
        label_names: Optional[List[str]] = None,
        edges: Optional[List[Tuple[int, int]]] = None,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param score_threshold: Confidence score threshold for detected keypoints.
        @type score_threshold: float
        @param label_names: Label names for the keypoints.
        @type label_names: Optional[List[str]]
        @param edges: Keypoint connection pairs for visualizing the skeleton. Example:
            [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint
            1, keypoint 1 is connected to keypoint 2, etc.
        @type edges: Optional[List[Tuple[int, int]]]
        """
        super().__init__(
            output_layer_name,
            score_threshold=score_threshold,
            label_names=label_names,
            edges=edges,
        )
        self._logger.debug(
            f"HRNetParser initialized with output_layer_name='{output_layer_name}', score_threshold={score_threshold}, label_names={label_names}, edges={edges}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "HRNetParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: HRNetParser
        """

        super().build(head_config)
        self._logger.debug("HRNetParser built successfully")
        return self

    def run(self):
        self._logger.debug("HRNetParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            heatmaps = output.getTensor(
                self.output_layer_name,
                dai.TensorInfo.StorageOrder.NCHW,  # this signals DAI to return output as NCHW
                dequantize=True,
            )

            if heatmaps.shape[0] == 1:
                heatmaps = heatmaps[0]  # remove batch dimension

            if len(heatmaps.shape) != 3:
                raise ValueError(
                    f"Expected 3D output tensor, got {len(heatmaps.shape)}D."
                )

            self.n_keypoints, map_h, map_w = heatmaps.shape

            scores = np.array([np.max(heatmap) for heatmap in heatmaps])
            scores = np.clip(scores, 0, 1)  # TODO: check why scores are sometimes >1

            keypoints = np.array(
                [
                    np.unravel_index(heatmap.argmax(), heatmap.shape)
                    for heatmap in heatmaps
                ]
            )
            keypoints = keypoints.astype(np.float32)
            keypoints = keypoints[:, ::-1] / np.array(
                [map_w, map_h]
            )  # normalize keypoints to [0, 1]

            keypoints_message = create_keypoints_message(
                keypoints=keypoints,
                scores=scores,
                confidence_threshold=self.score_threshold,
                edges=self.edges,
                label_names=self.label_names,
            )
            keypoints_message.setTimestamp(output.getTimestamp())
            keypoints_message.setTransformation(output.getTransformation())
            keypoints_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created keypoints message with {len(keypoints)} points"
            )

            self.out.send(keypoints_message)

            self._logger.debug("Keypoints message sent successfully")



import time

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.node import ImgFrameOverlay
from tests.utils import LOG_INTERVAL, OutputMock, create_img_frame

HEIGHT, WIDTH = 5, 5
ALPHA = 0.5
FRAME1 = np.random.randint(0, 255, (HEIGHT, WIDTH, 3), dtype=np.uint8)
FRAME2 = np.random.randint(0, 255, (HEIGHT, WIDTH, 3), dtype=np.uint8)
FRAME3 = np.random.randint(0, 255, (HEIGHT, WIDTH * 2, 3), dtype=np.uint8)


@pytest.fixture(scope="session")
def duration(request):
    d = request.config.getoption("--duration")
    if d is None:
        return 1e-6
    return d


@pytest.fixture
def overlayer():
    return ImgFrameOverlay()


def test_building(overlayer: ImgFrameOverlay):
    overlayer.build(OutputMock(), OutputMock(), alpha=ALPHA)
    assert overlayer._alpha == ALPHA


def test_parameter_setting(overlayer: ImgFrameOverlay):
    overlayer.setAlpha(ALPHA)
    assert overlayer._alpha == ALPHA

    # test isinstance(alpha, float)
    with pytest.raises(ValueError):
        overlayer.setAlpha("not a float")

    # test 0.0 <= alpha <= 1.0
    with pytest.raises(ValueError):
        overlayer.setAlpha(ALPHA * 100)


def test_processing(
    overlayer: ImgFrameOverlay,
    duration: float,
):
    img_frame1: dai.ImgFrame = create_img_frame(FRAME1)
    img_frame2: dai.ImgFrame = create_img_frame(FRAME2)
    img_frame3: dai.ImgFrame = create_img_frame(FRAME3)

    o_background = OutputMock()
    o_foreground = OutputMock()
    overlayer.build(o_background, o_foreground)
    overlayer.setAlpha(ALPHA)

    q_background = o_background.createOutputQueue()
    q_foreground = o_foreground.createOutputQueue()
    q_overlayer = overlayer.out.createOutputQueue()

    start_time = time.time()
    last_log_time = time.time()
    while time.time() - start_time < duration:
        if time.time() - last_log_time > LOG_INTERVAL:
            print(
                f"Test running... {time.time()-start_time:.1f}s elapsed, {duration-time.time()+start_time:.1f}s remaining"
            )
            last_log_time = time.time()
        q_background.send(img_frame1)
        q_foreground.send(img_frame2)
        overlayer.process(q_background.get(), q_foreground.get())
        overlay = q_overlayer.get()

        assert isinstance(overlay, dai.ImgFrame)
        assert np.array_equal(
            overlay.getCvFrame(),
            np.round(FRAME1 * ALPHA + FRAME2 * (1 - ALPHA)).astype(np.uint8),
        )

        q_background.send(img_frame1)
        q_foreground.send(img_frame3)
        overlayer.process(q_background.get(), q_foreground.get())
        overlay = q_overlayer.get()

        assert overlay.getCvFrame().shape == img_frame1.getCvFrame().shape



from itertools import (
    product,  # NOTE: you can use manual_product function instead of itertools.product
)
from typing import List, Optional, Tuple

import numpy as np

from depthai_nodes.node.parsers.utils import normalize_bboxes
from depthai_nodes.node.parsers.utils.keypoints import normalize_keypoints


def manual_product(*args):
    """You can use this function instead of itertools.product."""
    if not args:
        return [()]
    result = [[]]
    for pool in args:
        result = [x + [y] for x in result for y in pool]
    return [tuple(x) for x in result]


def generate_anchors(
    input_size: Tuple[int, int],
    min_sizes: Optional[List[List[int]]] = None,
    strides: Optional[List[int]] = None,
):
    """Generate a set of default bounding boxes, known as anchors.
    The code is taken from https://github.com/Kazuhito00/YuNet-ONNX-TFLite-Sample/tree/main

    @param input_size: A tuple representing the width and height of the input image.
    @type input_size: Tuple[int, int]
    @param min_sizes: A list of lists, where each inner list contains the minimum sizes of the anchors for different feature maps. If None then '[[10, 16, 24], [32, 48], [64, 96], [128, 192, 256]]' will be used. Defaults to None.
    @type min_sizes Optional[List[List[int]]]
    @param strides: Strides for each feature map layer. If None then '[8, 16, 32, 64]' will be used. Defaults to None.
    @type strides: Optional[List[int]]
    @return: Anchors.
    @rtype: np.ndarray
    """
    w, h = input_size

    if min_sizes is None:
        min_sizes = [[10, 16, 24], [32, 48], [64, 96], [128, 192, 256]]
    if strides is None:
        strides = [8, 16, 32, 64]

    # Calculate sizes of different feature maps by progressively halving the dimensions of the input image.
    feature_map_2th = [int(int((h + 1) / 2) / 2), int(int((w + 1) / 2) / 2)]
    feature_map_3th = [int(feature_map_2th[0] / 2), int(feature_map_2th[1] / 2)]
    feature_map_4th = [int(feature_map_3th[0] / 2), int(feature_map_3th[1] / 2)]
    feature_map_5th = [int(feature_map_4th[0] / 2), int(feature_map_4th[1] / 2)]
    feature_map_6th = [int(feature_map_5th[0] / 2), int(feature_map_5th[1] / 2)]
    feature_maps = [feature_map_3th, feature_map_4th, feature_map_5th, feature_map_6th]

    # Generate anchors
    anchors = []
    for k, f in enumerate(feature_maps):
        min_sizes_k = min_sizes[k]
        for i, j in product(range(f[0]), range(f[1])):
            for min_size_k in min_sizes_k:
                s_kx = min_size_k / w
                s_ky = min_size_k / h

                cx = (j + 0.5) * strides[k] / w
                cy = (i + 0.5) * strides[k] / h

                anchors.append([cx, cy, s_kx, s_ky])

    anchors = np.array(anchors, dtype=np.float32)
    return anchors


def decode_detections(
    input_size: Tuple[int, int],
    loc: np.ndarray,
    conf: np.ndarray,
    iou: np.ndarray,
    variance: Optional[List[float]] = None,
):
    """
    Decodes the output of an object detection model by converting the model's predictions (localization, confidence, and IoU scores) into bounding boxes, keypoints, and scores.
    The code is taken from https://github.com/Kazuhito00/YuNet-ONNX-TFLite-Sample/tree/main

    @param input_size: The size of the input image (height, width).
    @type input_size: tuple
    @param loc: The predicted locations (or offsets) of the bounding boxes.
    @type loc: np.ndarray
    @param conf: The predicted class confidence scores.
    @type conf: np.ndarray
    @param iou: The predicted IoU (Intersection over Union) scores.
    @type iou: np.ndarray
    @param variance: A list of variances used to decode the bounding box predictions. If None then [0.1,0.2] will be used. Defaults to None.
    @type variance: Optional[List[float]]
    @return: A tuple of bboxes, keypoints, and scores.
        - bboxes: NumPy array of shape (N, 4) containing the decoded bounding boxes in the format [x_min, y_min, width, height].
        - keypoints: A NumPy array of shape (N, 10) containing the decoded keypoint coordinates for each anchor.
        - scores: A NumPy array of shape (N, 1) containing the combined scores for each anchor.
    @rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]

    """

    w, h = input_size

    if variance is None:
        variance = [0.1, 0.2]

    anchors = generate_anchors(input_size)

    # Get scores
    cls_scores = conf[:, 1]
    iou_scores = iou[:, 0]
    _idx = np.where(iou_scores < 0.0)
    iou_scores[_idx] = 0.0
    _idx = np.where(iou_scores > 1.0)
    iou_scores[_idx] = 1.0
    scores = np.sqrt(cls_scores * iou_scores)
    scores = scores[:, np.newaxis]

    # Get bounding boxes
    scale = np.array((w, h))
    bboxes = np.hstack(
        (
            (anchors[:, 0:2] + loc[:, 0:2] * variance[0] * anchors[:, 2:4]) * scale,
            (anchors[:, 2:4] * np.exp(loc[:, 2:4] * variance)) * scale,
        )
    )
    bboxes[:, 0:2] -= bboxes[:, 2:4] / 2

    # Get keypoints
    keypoints = np.hstack(
        (
            (anchors[:, 0:2] + loc[:, 4:6] * variance[0] * anchors[:, 2:4]) * scale,
            (anchors[:, 0:2] + loc[:, 6:8] * variance[0] * anchors[:, 2:4]) * scale,
            (anchors[:, 0:2] + loc[:, 8:10] * variance[0] * anchors[:, 2:4]) * scale,
            (anchors[:, 0:2] + loc[:, 10:12] * variance[0] * anchors[:, 2:4]) * scale,
            (anchors[:, 0:2] + loc[:, 12:14] * variance[0] * anchors[:, 2:4]) * scale,
        )
    )

    return bboxes, keypoints, scores


def prune_detections(
    bboxes: np.ndarray, keypoints: np.ndarray, scores: np.ndarray, conf_threshold: float
):
    """Prune detections based on confidence threshold.

    Parameters:
    @param bboxes: A numpy array of shape (N, 4) containing the bounding boxes.
    @type np.ndarray
    @param keypoints: A numpy array of shape (N, 10) containing the keypoints.
    @type np.ndarray
    @param scores: A numpy array of shape (N,) containing the scores.
    @type np.ndarray
    @param conf_threshold: The confidence threshold.
    @type float
    @return: A tuple of bboxes, keypoints, and scores.
        - bboxes: NumPy array of shape (N, 4) containing the decoded bounding boxes in the format [x_min, y_min, width, height].
        - keypoints: A NumPy array of shape (N, 10) containing the decoded keypoint coordinates for each anchor.
        - scores: A NumPy array of shape (N, 1) containing the combined scores for each anchor.
    @rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]
    """

    keep_indices = np.where(scores.squeeze() > conf_threshold)
    return bboxes[keep_indices], keypoints[keep_indices], scores[keep_indices]


def format_detections(
    bboxes: np.ndarray,
    keypoints: np.ndarray,
    scores: np.ndarray,
    input_size: Tuple[int, int],
):
    """Format detections into a list of dictionaries.

    @param bboxes: A numpy array of shape (N, 4) containing the bounding boxes.
    @type np.ndarray
    @param keypoints: A numpy array of shape (N, 10) containing the keypoints.
    @type np.ndarray
    @param scores: A numpy array of shape (N,) containing the scores.
    @type np.ndarray
    @param input_size: A tuple representing the width and height of the input image.
    @type input_size: tuple
    @return: A tuple of bboxes, keypoints, and scores.
        - bboxes: NumPy array of shape (N, 4) containing the decoded bounding boxes in the format [x_min, y_min, width, height].
        - keypoints: A NumPy array of shape (N, 10) containing the decoded keypoint coordinates for each anchor.
        - scores: A NumPy array of shape (N, 1) containing the combined scores for each anchor.
    @rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]
    """

    w, h = input_size

    bboxes = normalize_bboxes(bboxes, height=h, width=w)

    keypoints = keypoints.astype(np.int32)
    keypoints = keypoints.reshape(-1, 2)
    keypoints = normalize_keypoints(keypoints, height=h, width=w)
    keypoints = keypoints.reshape(-1, 5, 2)  # (N,2) to (N,5,2)

    scores = np.array([scores]).flatten()

    return bboxes, keypoints, scores



from typing import Any, Dict, List, Tuple

import numpy as np


def local_maximum_filter(x: np.ndarray, kernel_size: int) -> np.ndarray:
    """Apply a local maximum filter to the input array.

    @param x: Input array.
    @type x: np.ndarray
    @param kernel_size: Size of the local maximum filter.
    @type kernel_size: int
    @return: Output array after applying the local maximum filter.
    @rtype: np.ndarray
    """
    # Ensure input is a 4D array (e.g., batch, channels, height, width)
    if len(x.shape) != 4:
        raise ValueError("Input array must be 4-dimensional.")

    if x.shape[0] != 1 and x.shape[1] != 1:
        raise ValueError("Batch size and number of channels must be 1.")

    _, _, height, width = x.shape

    # Pad the input array
    pad_width = kernel_size // 2
    padded_x = np.pad(
        x,
        ((0, 0), (0, 0), (pad_width, pad_width), (pad_width, pad_width)),
        mode="constant",
    )

    # Use stride tricks to generate a view of the array with sliding windows
    shape = (
        padded_x.shape[0],
        padded_x.shape[1],
        height,
        width,
        kernel_size,
        kernel_size,
    )
    strides = (
        padded_x.strides[0],
        padded_x.strides[1],
        padded_x.strides[2],
        padded_x.strides[3],
        padded_x.strides[2],
        padded_x.strides[3],
    )

    sliding_window_view = np.lib.stride_tricks.as_strided(
        padded_x, shape=shape, strides=strides
    )

    # Compute the local maximum over the sliding windows
    local_max = np.max(sliding_window_view, axis=(4, 5))

    return local_max


def normgrid(x, H, W):
    """Normalize coords to [-1,1].

    @param x: Input coordinates, shape (N, Hg, Wg, 2)
    @type x: np.ndarray
    @param H: Height of the output feature map
    @type H: int
    @param W: Width of the output feature map
    @type W: int
    @return: Normalized coordinates, shape (N, Hg, Wg, 2)
    @rtype: np.ndarray
    """
    return 2.0 * (x / np.array([W - 1, H - 1], dtype=x.dtype)) - 1.0


def bilinear(im, pos, H, W):
    """Given an input and a flow-field grid, computes the output using input values and
    pixel locations from grid. Supported only bilinear interpolation method to sample
    the input pixels.

    @param im: Input feature map, shape (N, C, H, W)
    @type im: np.ndarray
    @param pos: Point coordinates, shape (N, Hg, Wg, 2)
    @type pos: np.ndarray
    @param H: Height of the output feature map
    @type H: int
    @param W: Width of the output feature map
    @type W: int
    @return: A tensor with sampled points, shape (N, C, Hg, Wg)
    @rtype: np.ndarray
    """
    align_corners = False
    n, c, h, w = im.shape
    grid = normgrid(pos, H, W)[..., np.newaxis]
    grid = grid.transpose(0, 1, 3, 2)

    x = grid[..., 0]
    y = grid[..., 1]

    if align_corners:
        x = ((x + 1) / 2) * (w - 1)
        y = ((y + 1) / 2) * (h - 1)
    else:
        x = ((x + 1) * w - 1) / 2
        y = ((y + 1) * h - 1) / 2

    x = x.reshape(n, -1)
    y = y.reshape(n, -1)

    x0 = np.floor(x).astype(np.int32)
    y0 = np.floor(y).astype(np.int32)
    x1 = x0 + 1
    y1 = y0 + 1

    wa = ((x1 - x) * (y1 - y))[:, np.newaxis]
    wb = ((x1 - x) * (y - y0))[:, np.newaxis]
    wc = ((x - x0) * (y1 - y))[:, np.newaxis]
    wd = ((x - x0) * (y - y0))[:, np.newaxis]

    # Apply padding
    im_padded = np.pad(
        im, ((0, 0), (0, 0), (1, 1), (1, 1)), mode="constant", constant_values=0
    )
    padded_h = h + 2
    padded_w = w + 2

    # Adjust points for padding
    x0, x1 = x0 + 1, x1 + 1
    y0, y1 = y0 + 1, y1 + 1

    # Clip coordinates to stay within bounds
    x0 = np.clip(x0, 0, padded_w - 1)
    x1 = np.clip(x1, 0, padded_w - 1)
    y0 = np.clip(y0, 0, padded_h - 1)
    y1 = np.clip(y1, 0, padded_h - 1)

    # Flatten im_padded for indexing
    im_padded = im_padded.reshape(n, c, -1)

    x0_y0 = (x0 + y0 * padded_w)[:, np.newaxis].repeat(c, axis=1)
    x0_y1 = (x0 + y1 * padded_w)[:, np.newaxis].repeat(c, axis=1)
    x1_y0 = (x1 + y0 * padded_w)[:, np.newaxis].repeat(c, axis=1)
    x1_y1 = (x1 + y1 * padded_w)[:, np.newaxis].repeat(c, axis=1)

    Ia = np.take_along_axis(im_padded, x0_y0, axis=2)
    Ib = np.take_along_axis(im_padded, x0_y1, axis=2)
    Ic = np.take_along_axis(im_padded, x1_y0, axis=2)
    Id = np.take_along_axis(im_padded, x1_y1, axis=2)

    result = (Ia * wa + Ib * wb + Ic * wc + Id * wd).reshape(
        n, c, grid.shape[1], grid.shape[2]
    )
    return result


def _get_kpts_heatmap(
    kpts: np.ndarray,
    softmax_temp: float = 1.0,
) -> np.ndarray:
    """Get the keypoints heatmap.

    @param kpts: Keypoints.
    @type kpts: np.ndarray
    @param softmax_temp: Softmax temperature.
    @type softmax_temp: float
    @return: Keypoints heatmap.
    @rtype: np.ndarray
    """
    kpts = np.exp(kpts * softmax_temp)
    scores = kpts / np.sum(kpts, axis=1, keepdims=True)
    scores = scores[:, :64]
    B, _, H, W = scores.shape
    heatmap = np.transpose(scores, (0, 2, 3, 1)).reshape(B, H, W, 8, 8)
    heatmap = np.transpose(heatmap, (0, 1, 3, 2, 4)).reshape(B, 1, H * 8, W * 8)
    return heatmap


def _nms(
    x: np.ndarray,
    threshold: float = 0.05,
    kernel_size: int = 5,
) -> np.ndarray:
    """Non-Maximum Suppression.

    @param x: Input array.
    @type x: np.ndarray
    @param threshold: Non-maximum suppression threshold.
    @type threshold: float
    @param kernel_size: Size of the local maximum filter.
    @type kernel_size: int
    @return: Output array after applying non-maximum suppression.
    @rtype: np.ndarray
    """
    # Non-Maximum Suppression
    B, _, H, W = x.shape
    local_max = local_maximum_filter(x, kernel_size)
    pos = (x == local_max) & (x > threshold)

    pos_batched = [np.fliplr(np.argwhere(k)[:, 1:]) for k in pos]

    pad_val = max(len(k) for k in pos_batched)
    pos_array = np.zeros((B, pad_val, 2), dtype=int)

    for b, kpts in enumerate(pos_batched):
        pos_array[b, : len(kpts)] = kpts

    return pos_array


def detect_and_compute(
    feats: np.ndarray,
    kpts: np.ndarray,
    heatmaps: np.ndarray,
    resize_rate_w: float,
    resize_rate_h: float,
    input_size: Tuple[int, int],
    top_k: int = 4096,
) -> List[Dict[str, Any]]:
    """Detect and compute keypoints.

    @param feats: Features.
    @type feats: np.ndarray
    @param kpts: Keypoints.
    @type kpts: np.ndarray
    @param resize_rate_w: Resize rate for width.
    @type resize_rate_w: float
    @param resize_rate_h: Resize rate for height.
    @type resize_rate_h: float
    @param input_size: Input size.
    @type input_size: Tuple[int, int]
    @param top_k: Maximum number of keypoints to keep.
    @type top_k: int
    @return: List of dictionaries containing keypoints, scores, and descriptors.
    @rtype: List[Dict[str, Any]]
    """
    norm = np.linalg.norm(feats, axis=1, keepdims=True)
    feats = feats / norm

    kpts_heats = _get_kpts_heatmap(kpts)
    mkpts = _nms(kpts_heats, threshold=0.05, kernel_size=5)  # int64

    if mkpts.size == 0:
        return None
    # Numpy implementation of F.grid_sample
    nearest_result = bilinear(kpts_heats, mkpts, input_size[1], input_size[0])
    bilinear_result = bilinear(heatmaps, mkpts, input_size[1], input_size[0])

    scores = (nearest_result * bilinear_result).reshape(1, -1)

    scores = scores.astype(np.float32)
    mkpts = mkpts.astype(np.int64)

    scores[np.all(mkpts == 0, axis=-1)] = -1

    idxs = np.argsort(-scores)
    mkpts_x = np.take_along_axis(mkpts[..., 0], idxs, axis=-1)[:, :top_k]
    mkpts_y = np.take_along_axis(mkpts[..., 1], idxs, axis=-1)[:, :top_k]
    mkpts = np.stack([mkpts_x, mkpts_y], axis=-1)
    mkpts = mkpts.astype(np.float32)

    scores = np.take_along_axis(scores, idxs, axis=-1)[:, :top_k]

    feats = bilinear(feats, mkpts, input_size[1], input_size[0])
    feats = feats[0].transpose(2, 1, 0)

    norm = np.linalg.norm(feats, axis=-1, keepdims=True)
    feats = feats / norm

    mkpts = mkpts.astype(np.float32)
    mkpts *= np.array([resize_rate_w, resize_rate_h])[None, None, :]

    valid = scores > 0
    result = []
    valid = valid[0]
    result.append(
        {
            "keypoints": mkpts[0][valid],
            "scores": scores[0][valid],
            "descriptors": feats[0][valid],
        }
    )

    return result


def _match_mkpts(
    feats1: np.ndarray, feats2: np.ndarray, min_cossim: float = 0.62
) -> Tuple[np.ndarray, np.ndarray]:
    """Match features.

    @param feats1: Features 1.
    @type feats1: np.ndarray
    @param feats2: Features 2.
    @type feats2: np.ndarray
    @param min_cossim: Minimum cosine similarity.
    @type min_cossim: float
    @return: Matched features.
    @rtype: Tuple[np.ndarray, np.ndarray]
    """
    cossim = feats1 @ feats2.T
    cossim_t = feats2 @ feats1.T
    match12 = np.argmax(cossim, axis=1)
    match21 = np.argmax(cossim_t, axis=1)

    idx0 = np.arange(len(match12))
    mutual = match21[match12] == idx0

    if min_cossim > 0:
        max_cossim = np.max(cossim, axis=1)
        good = max_cossim > min_cossim
        idx0 = idx0[mutual & good]
        idx1 = match12[mutual & good]
    else:
        idx0 = idx0[mutual]
        idx1 = match12[mutual]

    return idx0, idx1


def match(
    result1: Dict[str, Any], result2: Dict[str, Any], min_cossim: float = -1
) -> Tuple[np.ndarray, np.ndarray]:
    """Match keypoints.

    @param result1: Result 1.
    @type result1: Dict[str, Any]
    @param result2: Result 2.
    @type result2: Dict[str, Any]
    @param min_cossim: Minimum cosine similarity.
    @type min_cossim: float
    @return: Matched keypoints.
    @rtype: Tuple[np.ndarray, np.ndarray]
    """
    indexes1, indexes2 = _match_mkpts(
        result1["descriptors"],
        result2["descriptors"],
        min_cossim=min_cossim,
    )

    mkpts0 = result1["keypoints"][indexes1]
    mkpts1 = result2["keypoints"][indexes2]

    return mkpts0, mkpts1



import depthai as dai
import numpy as np
import pytest

from depthai_nodes import (
    ImgDetectionExtended,
    ImgDetectionsExtended,
    Keypoint,
    Keypoints,
)


@pytest.fixture
def img_detection_extended():
    return ImgDetectionExtended()


@pytest.fixture
def img_detections_extended():
    return ImgDetectionsExtended()


def test_img_detection_extended_initialization(
    img_detection_extended: ImgDetectionExtended,
):
    assert img_detection_extended.confidence == -1.0
    assert img_detection_extended.label == -1
    assert img_detection_extended.label_name == ""


def test_img_detection_extended_set_rotated_rect(
    img_detection_extended: ImgDetectionExtended,
):
    rect = (0.5, 0.5, 0.2, 0.2, 45.0)
    img_detection_extended.rotated_rect = rect
    assert np.allclose(img_detection_extended.rotated_rect.center.x, 0.5, atol=1e-3)
    assert np.allclose(img_detection_extended.rotated_rect.center.y, 0.5, atol=1e-3)
    assert np.allclose(img_detection_extended.rotated_rect.size.width, 0.2, atol=1e-3)
    assert np.allclose(img_detection_extended.rotated_rect.size.height, 0.2, atol=1e-3)
    assert np.allclose(img_detection_extended.rotated_rect.angle, 45.0, atol=1e-3)


def test_img_detection_extended_set_confidence(
    img_detection_extended: ImgDetectionExtended,
):
    img_detection_extended.confidence = 0.9
    assert img_detection_extended.confidence == 0.9

    with pytest.raises(TypeError):
        img_detection_extended.confidence = "not a float"

    with pytest.raises(ValueError):
        img_detection_extended.confidence = 1.5


def test_img_detection_extended_set_label(img_detection_extended: ImgDetectionExtended):
    img_detection_extended.label = 1
    assert img_detection_extended.label == 1

    with pytest.raises(TypeError):
        img_detection_extended.label = "not an int"


def test_img_detection_extended_set_label_name(
    img_detection_extended: ImgDetectionExtended,
):
    img_detection_extended.label_name = "cat"
    assert img_detection_extended.label_name == "cat"

    with pytest.raises(TypeError):
        img_detection_extended.label_name = 123


def test_img_detection_extended_set_keypoints(
    img_detection_extended: ImgDetectionExtended,
):
    keypoints = []
    kp1 = Keypoint()
    kp1.x = 0.1
    kp1.y = 0.2
    kp2 = Keypoint()
    kp2.x = 0.3
    kp2.y = 0.4
    keypoints.append(kp1)
    keypoints.append(kp2)
    kpts = Keypoints()
    kpts.keypoints = keypoints
    img_detection_extended.keypoints = kpts
    for i, kp in enumerate(img_detection_extended.keypoints):
        assert kp.x == keypoints[i].x
        assert kp.y == keypoints[i].y

    with pytest.raises(TypeError):
        img_detection_extended.keypoints = "not a Keypoints object"


def test_img_detections_extended_initialization(
    img_detections_extended: ImgDetectionsExtended,
):
    assert img_detections_extended.detections == []
    assert isinstance(img_detections_extended.masks, np.ndarray)
    assert img_detections_extended.transformation is None


def test_img_detections_extended_set_detections(
    img_detections_extended: ImgDetectionsExtended,
):
    detection1 = ImgDetectionExtended()
    detection2 = ImgDetectionExtended()
    detections_list = [detection1, detection2]
    img_detections_extended.detections = detections_list
    assert img_detections_extended.detections == detections_list

    with pytest.raises(TypeError):
        img_detections_extended.detections = "not a list"

    with pytest.raises(TypeError):
        img_detections_extended.detections = [detection1, "not an ImgDetectionExtended"]


def test_img_detections_extended_set_masks(
    img_detections_extended: ImgDetectionsExtended,
):
    masks = np.random.randint(0, 256, (480, 640), dtype=np.int16)
    img_detections_extended.masks = masks
    assert np.array_equal(img_detections_extended.masks, masks)

    with pytest.raises(TypeError):
        img_detections_extended.masks = "not a numpy array"

    with pytest.raises(ValueError):
        img_detections_extended.masks = np.random.randint(
            0, 256, (480, 640, 3), dtype=np.int16
        )

    with pytest.raises(ValueError):
        img_detections_extended.masks = np.random.randint(
            0, 256, (480, 640), dtype=np.uint8
        )

    with pytest.raises(ValueError):
        img_detections_extended.masks = np.random.randint(
            -2, 256, (480, 640), dtype=np.int16
        )


def test_img_detections_extended_set_transformation(
    img_detections_extended: ImgDetectionsExtended,
):
    transformation = dai.ImgTransformation()
    img_detections_extended.transformation = transformation
    assert img_detections_extended.transformation == transformation

    with pytest.raises(TypeError):
        img_detections_extended.transformation = "not a dai.ImgTransformation"


def test_img_detections_extended_set_transformation_none(
    img_detections_extended: ImgDetectionsExtended,
):
    img_detections_extended.transformation = None
    assert img_detections_extended.transformation is None



from .apply_colormap import ApplyColormap
from .base_host_node import BaseHostNode
from .depth_merger import DepthMerger
from .gather_data import GatherData
from .host_parsing_neural_network import HostParsingNeuralNetwork
from .host_spatials_calc import HostSpatialsCalc
from .img_detections_bridge import ImgDetectionsBridge
from .img_detections_filter import ImgDetectionsFilter
from .img_frame_overlay import ImgFrameOverlay
from .parser_generator import ParserGenerator
from .parsers.base_parser import BaseParser
from .parsers.classification import ClassificationParser
from .parsers.classification_sequence import ClassificationSequenceParser
from .parsers.detection import DetectionParser
from .parsers.embeddings import EmbeddingsParser
from .parsers.fastsam import FastSAMParser
from .parsers.hrnet import HRNetParser
from .parsers.image_output import ImageOutputParser
from .parsers.keypoints import KeypointParser
from .parsers.lane_detection import LaneDetectionParser
from .parsers.map_output import MapOutputParser
from .parsers.mediapipe_palm_detection import MPPalmDetectionParser
from .parsers.mlsd import MLSDParser
from .parsers.ppdet import PPTextDetectionParser
from .parsers.regression import RegressionParser
from .parsers.scrfd import SCRFDParser
from .parsers.segmentation import SegmentationParser
from .parsers.superanimal_landmarker import SuperAnimalParser
from .parsers.xfeat import XFeatMonoParser, XFeatStereoParser
from .parsers.yolo import YOLOExtendedParser
from .parsers.yunet import YuNetParser
from .parsing_neural_network import ParsingNeuralNetwork
from .snaps_producer import SnapsProducer, SnapsProducerFrameOnly
from .tiles_patcher import TilesPatcher
from .tiling import Tiling

__all__ = [
    "ApplyColormap",
    "DepthMerger",
    "Tiling",
    "TilesPatcher",
    "ParserGenerator",
    "ParsingNeuralNetwork",
    "HostParsingNeuralNetwork",
    "HostSpatialsCalc",
    "ImageOutputParser",
    "YuNetParser",
    "MPPalmDetectionParser",
    "SCRFDParser",
    "SegmentationParser",
    "SuperAnimalParser",
    "KeypointParser",
    "MLSDParser",
    "XFeatMonoParser",
    "XFeatStereoParser",
    "ClassificationParser",
    "YOLOExtendedParser",
    "FastSAMParser",
    "RegressionParser",
    "HRNetParser",
    "PPTextDetectionParser",
    "MapOutputParser",
    "ClassificationSequenceParser",
    "LaneDetectionParser",
    "BaseParser",
    "DetectionParser",
    "EmbeddingsParser",
    "GatherData",
    "ImgFrameOverlay",
    "ImgDetectionsBridge",
    "ImgDetectionsFilter",
    "SnapsProducerFrameOnly",
    "SnapsProducer",
    "BaseHostNode",
]



import copy
from typing import List

import depthai as dai
import numpy as np
from numpy.typing import NDArray

from depthai_nodes import PRIMARY_COLOR, SECONDARY_COLOR


class Classifications(dai.Buffer):
    """Classification class for storing the classes and their respective scores.

    Attributes
    ----------
    classes : list[str]
        A list of classes.
    scores : NDArray[np.float32]
        Corresponding probability scores.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self):
        """Initializes the Classifications object."""
        dai.Buffer.__init__(self)
        self._classes: List[str] = []
        self._scores: NDArray[np.float32] = np.array([])
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the Classifications class and copies the
        attributes.

        @return: A new instance of the Classifications class.
        @rtype: Classifications
        """
        new_obj = Classifications()
        new_obj.classes = copy.deepcopy(self.classes)
        new_obj.scores = copy.deepcopy(self.scores)
        new_obj.setSequenceNum(self.getSequenceNum())
        new_obj.setTimestamp(self.getTimestamp())
        new_obj.setTimestampDevice(self.getTimestampDevice())
        new_obj.setTransformation(self.transformation)
        return new_obj

    @property
    def classes(self) -> List:
        """Returns the list of classes.

        @return: List of classes.
        @rtype: List[str]
        """
        return self._classes

    @classes.setter
    def classes(self, value: List[str]):
        """Sets the classes.

        @param value: A list of class names.
        @type value: List[str]
        @raise TypeError: If value is not a list.
        @raise ValueError: If each element is not of type string.
        """
        if not isinstance(value, List):
            raise TypeError(f"Classes must be a list, instead got {type(value)}.")
        if not all(isinstance(class_name, str) for class_name in value):
            raise ValueError("Classes must be a list of strings.")
        self._classes = value

    @property
    def scores(self) -> NDArray:
        """Returns the list of scores.

        @return: List of scores.
        @rtype: NDArray[np.float32]
        """
        return self._scores

    @scores.setter
    def scores(self, value: NDArray[np.float32]):
        """Sets the scores.

        @param value: A list of scores.
        @type value: NDArray[np.float32]
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 1D numpy array.
        @raise ValueError: If each element is not of type float.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError(f"Scores must be a np.ndarray, instead got {type(value)}.")
        if value.ndim != 1:
            raise ValueError("Scores must be a 1D a np.ndarray.")
        if value.size > 0 and value.dtype != np.float32:
            raise ValueError("Scores must be a np.ndarray of floats.")
        self._scores = value

    @property
    def top_class(self) -> str:
        """Returns the most probable class. Only works if classes are sorted by scores.

        @return: The top class.
        @rtype: str
        """
        return self._classes[0]

    @property
    def top_score(self) -> float:
        """Returns the probability of the most probable class. Only works if scores are
        sorted by descending order.

        @return: The top score.
        @rtype: float
        """
        return self._scores[0]

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """

        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )
        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        self.transformation = transformation

    def getVisualizationMessage(self) -> dai.ImgAnnotations:
        """Returns default visualization message for classification.

        The message adds the top five classes and their scores to the right side of the
        image.
        """
        img_annotations = dai.ImgAnnotations()
        annotation = dai.ImgAnnotation()
        w, h = self.transformation.getSize()

        font_size = h / 30
        x_offset = 2 / w
        y_offset = 2 / h

        for i in range(min(5, len(self._classes))):
            text = dai.TextAnnotation()
            text.position = dai.Point2f(
                x_offset,
                y_offset + (font_size / h) + i * (font_size / h),
                normalized=True,
            )
            text.text = f"{self._classes[i]} {self._scores[i] * 100:.0f}%"
            text.fontSize = font_size
            text.textColor = PRIMARY_COLOR if i == 0 else SECONDARY_COLOR
            annotation.texts.append(text)

        img_annotations.annotations.append(annotation)
        img_annotations.setTimestamp(self.getTimestamp())
        return img_annotations



from .queue import InfiniteQueueMock, QueueMock


class InputMock:
    """Input class to simulate the depthai input node."""

    def __init__(self):
        self._queue = QueueMock()

    def get(self):
        return self._queue.get()

    def tryGet(self):
        return self._queue.tryGet()

    def send(self, message):
        self._queue.send(message)


class InfiniteInputMock(InputMock):
    """Special input class that uses InfiniteQueue to simulate a long running test."""

    def __init__(self):
        self._queue = InfiniteQueueMock()



from .classification import (
    create_classification_message,
    create_classification_sequence_message,
)
from .clusters import create_cluster_message
from .detection import (
    create_detection_message,
)
from .image import create_image_message
from .keypoints import create_keypoints_message
from .line import create_line_detection_message
from .map import create_map_message
from .regression import create_regression_message
from .segmentation import create_segmentation_message
from .tracked_features import create_tracked_features_message

__all__ = [
    "create_image_message",
    "create_segmentation_message",
    "create_detection_message",
    "create_line_detection_message",
    "create_tracked_features_message",
    "create_keypoints_message",
    "create_classification_message",
    "create_regression_message",
    "create_map_message",
    "create_classification_sequence_message",
    "create_cluster_message",
]



from typing import Any, Dict

import depthai as dai

from depthai_nodes.node.parsers.base_parser import BaseParser


class EmbeddingsParser(BaseParser):
    """Parser class for parsing the output of embeddings neural network model head.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.

    Output Message/s
    ----------------
    **Type**: dai.NNData

    **Description**: The output layer of the neural network model head.
    """

    def __init__(self) -> None:
        """Initialize the EmbeddingsParser node."""
        super().__init__()
        self.output_layer_name: str = None
        self._logger.debug(
            f"EmbeddingsParser initialized with output_layer_name={self.output_layer_name}"
        )

    def setOutputLayerNames(self, output_layer_name: str) -> None:
        """Sets the output layer name for the parser.

        @param output_layer_name: The output layer name for the parser.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")

        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to {self.output_layer_name}")

    def build(self, head_config: Dict[str, Any]) -> "EmbeddingsParser":
        """Sets the head configuration for the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: EmbeddingsParser
        """

        self.output_layer_name = head_config["outputs"]
        assert (
            len(self.output_layer_name) == 1
        ), "Embeddings head should have only one output layer"

        self._logger.debug(
            f"EmbeddingsParser built with output_layer_name={self.output_layer_name}"
        )

        return self

    def run(self):
        self._logger.debug("EmbeddingsParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data

            # Get all the layer names
            output_names = self.output_layer_name or output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {output_names}")

            assert (
                len(output_names) == 1
            ), "Embeddings head should have only one output layer"
            output.setSequenceNum(output.getSequenceNum())

            self.out.send(output)

            self._logger.debug("Message sent successfully")



import time
from typing import Union

import depthai as dai
import numpy as np
import pytest
from pytest import FixtureRequest

from depthai_nodes import ImgDetectionExtended, ImgDetectionsExtended
from depthai_nodes.node.depth_merger import DepthMerger
from depthai_nodes.node.host_spatials_calc import HostSpatialsCalc
from tests.utils import (
    LOG_INTERVAL,
    OutputMock,
    create_img_detection,
    create_img_detection_extended,
    create_img_detections,
    create_img_detections_extended,
)

from .utils.calibration_handler import get_calibration_handler

DIFFERENT_TESTS = 4  # used for stability tests


@pytest.fixture(scope="session")
def duration(request):
    d = request.config.getoption("--duration")
    if d is None:
        return 1e-6
    return d


@pytest.fixture
def depth_merger():
    depth_merger = DepthMerger(shrinking_factor=0.0)
    calib_handler = get_calibration_handler()
    depth_merger.host_spatials_calc = HostSpatialsCalc(
        calib_data=calib_handler, depth_alignment_socket=dai.CameraBoardSocket.CAM_A
    )
    return depth_merger


@pytest.fixture
def img_detection():
    return create_img_detection()


@pytest.fixture
def img_detection_extended():
    return create_img_detection_extended()


@pytest.fixture
def depth_frame():
    frame = np.random.randint(200, 30000, (720, 1280), dtype=np.uint16)
    img_frame = dai.ImgFrame()
    img_frame.setCvFrame(frame, dai.ImgFrame.Type.RAW16)
    img_frame.setWidth(1280)
    img_frame.setHeight(720)
    return img_frame


@pytest.fixture
def img_detections():
    return create_img_detections()


@pytest.fixture
def img_detections_extended():
    return create_img_detections_extended()


def verify_spatial_detection(spatial_det, img_detection):
    assert isinstance(spatial_det, dai.SpatialImgDetection)

    if isinstance(img_detection, ImgDetectionExtended):
        xmin, ymin, xmax, ymax = img_detection.rotated_rect.getOuterRect()
    else:
        xmin = img_detection.xmin
        ymin = img_detection.ymin
        xmax = img_detection.xmax
        ymax = img_detection.ymax
    np.testing.assert_almost_equal(spatial_det.xmin, xmin, decimal=2)
    np.testing.assert_almost_equal(spatial_det.ymin, ymin, decimal=2)
    np.testing.assert_almost_equal(spatial_det.xmax, xmax, decimal=2)
    np.testing.assert_almost_equal(spatial_det.ymax, ymax, decimal=2)
    assert spatial_det.label == img_detection.label
    np.testing.assert_almost_equal(
        spatial_det.confidence, img_detection.confidence, decimal=2
    )


def test_initialization(depth_merger: DepthMerger):
    assert depth_merger.shrinking_factor == 0.0


@pytest.mark.parametrize("detection", ["img_detection", "img_detection_extended"])
def test_img_detection(
    duration: float,
    depth_merger: DepthMerger,
    depth_frame: dai.ImgFrame,
    request: FixtureRequest,
    detection: str,
):
    img_detection: Union[
        ImgDetectionExtended, dai.ImgDetection
    ] = request.getfixturevalue(detection)
    output_2d = OutputMock()
    output_depth = OutputMock()

    modified_duration = duration / DIFFERENT_TESTS

    depth_merger.build(
        output_2d=output_2d,
        output_depth=output_depth,
        calib_data=get_calibration_handler(),
        depth_alignment_socket=dai.CameraBoardSocket.CAM_A,
        shrinking_factor=0.0,
    )

    q_2d = output_2d.createOutputQueue()
    q_depth = output_depth.createOutputQueue()
    q_out = depth_merger.output.createOutputQueue()

    start_time = time.time()
    last_log_time = time.time()
    while time.time() - start_time < modified_duration:
        if time.time() - last_log_time > LOG_INTERVAL:
            print(
                f"Test running... {time.time()-start_time:.1f}s elapsed, {modified_duration-time.time()+start_time:.1f}s remaining"
            )
            last_log_time = time.time()
        output_2d.send(img_detection)
        output_depth.send(depth_frame)

        depth_merger.process(q_2d.get(), q_depth.get())

        spatial_det: dai.SpatialImgDetection = q_out.get()
        verify_spatial_detection(spatial_det, img_detection)


@pytest.mark.parametrize("detections", ["img_detections", "img_detections_extended"])
def test_img_detections(
    depth_merger: DepthMerger,
    depth_frame: dai.ImgFrame,
    request: FixtureRequest,
    detections: str,
    duration: float,
):
    img_detections: Union[
        ImgDetectionsExtended, dai.ImgDetections
    ] = request.getfixturevalue(detections)
    output_2d = OutputMock()
    output_depth = OutputMock()

    modified_duration = duration / DIFFERENT_TESTS

    depth_merger.build(
        output_2d=output_2d,
        output_depth=output_depth,
        calib_data=get_calibration_handler(),
        depth_alignment_socket=dai.CameraBoardSocket.CAM_A,
        shrinking_factor=0.0,
    )

    q_2d = output_2d.createOutputQueue()
    q_depth = output_depth.createOutputQueue()
    q_out = depth_merger.output.createOutputQueue()

    start_time = time.time()
    last_log_time = time.time()
    while time.time() - start_time < modified_duration:
        if time.time() - last_log_time > LOG_INTERVAL:
            print(
                f"Test running... {time.time()-start_time:.1f}s elapsed, {modified_duration-time.time()+start_time:.1f}s remaining"
            )
            last_log_time = time.time()
        output_2d.send(img_detections)
        output_depth.send(depth_frame)

        depth_merger.process(q_2d.get(), q_depth.get())

        spatial_dets: dai.SpatialImgDetections = q_out.get()
        assert isinstance(spatial_dets, dai.SpatialImgDetections)
        assert len(spatial_dets.detections) == len(img_detections.detections)

        for i, spatial_det in enumerate(spatial_dets.detections):
            img_det = img_detections.detections[i]
            verify_spatial_detection(spatial_det, img_det)



import depthai as dai
import numpy as np

import depthai_nodes.message.creators as creators

from .constants import ARRAYS


def create_img_frame(
    image: np.ndarray = ARRAYS["3d"],
    img_frame_type=dai.ImgFrame.Type.BGR888p,
):
    return creators.create_image_message(
        image=image.astype(np.uint8), img_frame_type=img_frame_type
    )


def create_map(map: np.ndarray = ARRAYS["2d"]):
    return creators.create_map_message(map=map.astype(np.float32))


def create_segmentation(mask: np.ndarray = ARRAYS["2d"]):
    return creators.create_segmentation_message(mask=mask.astype(np.int16))



from typing import Any, Dict

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_regression_message
from depthai_nodes.node.parsers.base_parser import BaseParser


class RegressionParser(BaseParser):
    """Parser class for parsing the output of a model with regression output (e.g. Age-
    Gender).

    Attributes
    ----------
    output_layer_name : str
        Name of the output layer relevant to the parser.

    Output Message/s
    ----------------
    **Type**: Predictions

    **Description**: Message containing the prediction(s).
    """

    def __init__(
        self,
        output_layer_name: str = "",
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name : str
        """
        super().__init__()
        self.output_layer_name = output_layer_name
        self._logger.debug(
            f"RegressionParser initialized with output_layer_name='{output_layer_name}'"
        )

    def setOutputLayerName(self, output_layer_name: str):
        """Sets the name of the output layer.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "RegressionParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: RegressionParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for Regression, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]

        self._logger.debug(
            f"RegressionParser built with output_layer_name='{self.output_layer_name}'"
        )

        return self

    def run(self):
        self._logger.debug("RegressionParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            predictions = output.getTensor(
                self.output_layer_name, dequantize=True
            ).squeeze()
            predictions = np.atleast_1d(predictions).tolist()

            regression_message = create_regression_message(predictions=predictions)
            regression_message.setTimestamp(output.getTimestamp())
            regression_message.setTransformation(output.getTransformation())
            regression_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created regression message with {len(predictions)} values"
            )

            self.out.send(regression_message)

            self._logger.debug("Regression message sent successfully")



import time
from queue import PriorityQueue
from typing import (
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    TypeVar,
    runtime_checkable,
)

import depthai as dai

from depthai_nodes import GatheredData
from depthai_nodes.logging import get_logger


@runtime_checkable
class HasDetections(Protocol):
    @property
    def detections(self) -> List:
        ...


TReference = TypeVar("TReference", bound=dai.Buffer)
TGathered = TypeVar("TGathered", bound=dai.Buffer)


class GatherData(dai.node.ThreadedHostNode, Generic[TReference, TGathered]):
    """A class for gathering data. Gathers n messages based on reference_data. To
    determine n, wait_count_fn function is used. The default wait_count_fn function is
    waiting for len(TReference.detection). This means the node works out-of-the-box with
    dai.ImgDetections and ImgDetectionsExtended.

    Attributes
    ----------
    FPS_TOLERANCE_DIVISOR: float
        Divisor for the FPS tolerance.
    INPUT_CHECKS_PER_FPS: int
        Number of input checks per FPS.
    input_data: dai.Node.Input
        Input to be gathered.
    input_reference: dai.Node.Input
        Input to determine how many gathered items to wait for.
    output: dai.Node.Output
        Output for gathered data.
    """

    FPS_TOLERANCE_DIVISOR = 2.0
    INPUT_CHECKS_PER_FPS = 100

    def __init__(self) -> None:
        """Initializes the GatherData node."""
        super().__init__()
        self._camera_fps: Optional[int] = None
        self._unmatched_data: List[TGathered] = []
        self._data_by_reference_ts: Dict[float, List[TGathered]] = {}
        self._reference_data: Dict[float, TReference] = {}
        self._ready_timestamps = PriorityQueue()
        self._wait_count_fn = self._default_wait_count_fn

        self.input_data = self.createInput()
        self.input_reference = self.createInput()
        self.out = self.createOutput()

        self._logger = get_logger(__name__)
        self._logger.debug("GatherData initialized")

    @staticmethod
    def _default_wait_count_fn(reference: TReference) -> int:
        assert isinstance(reference, HasDetections)
        return len(reference.detections)

    def build(
        self,
        camera_fps: int,
        wait_count_fn: Optional[Callable[[TReference], int]] = None,
    ) -> "GatherData[TReference, TGathered]":
        """Builds and configures the GatherData node with the specified parameters.

        @param camera_fps: The frames per second (FPS) setting for the camera.
        @param wait_count_fn: A function that takes a reference and returns how many frames to wait.
        @type camera_fps: int
        @type wait_count_fn: Optional[Callable[[TReference], int]]

        @return: The configured GatherData node instance.
        @rtype: GatherData[TReference, TGathered]

        @example:
            >>> gather_node = GatherData()
            >>> gather_node.build(camera_fps=30)
            >>> def custom_wait(ref): return 2
            >>> gather_node.build(camera_fps=60, wait_count_fn=custom_wait)
        """
        self.set_camera_fps(camera_fps)
        if wait_count_fn is None:
            wait_count_fn = self._default_wait_count_fn
        self.set_wait_count_fn(wait_count_fn)
        self._logger.debug(f"GatherData built with camera_fps={camera_fps}")
        return self

    def set_camera_fps(self, fps: int) -> None:
        if fps <= 0:
            raise ValueError(f"Camera FPS must be positive, got {fps}")
        self._camera_fps = fps
        self._logger.debug(f"Camera FPS set to {fps}")

    def run(self) -> None:
        self._logger.debug("GatherData run started")
        if not self._camera_fps:
            raise ValueError("Camera FPS not set. Call build() before run().")

        while self.isRunning():
            try:
                input_data: TGathered = self.input_data.tryGet()
                input_reference: TReference = self.input_reference.tryGet()
            except dai.MessageQueue.QueueException:
                break
            if input_data:
                self._logger.debug("Input data received")
                self._add_data(input_data)
                self._send_ready_data()
            if input_reference:
                self._logger.debug("Input reference received")
                self._add_reference(input_reference)
                self._send_ready_data()

            time.sleep(1 / self.INPUT_CHECKS_PER_FPS / self._camera_fps)

    def _send_ready_data(self) -> None:
        ready_data = self._pop_ready_data()
        if ready_data:
            self._clear_old_data(ready_data)
            self.out.send(ready_data)
            self._logger.debug("Gathered data sent")

    def _add_data(self, data: TGathered) -> None:
        data_ts = self._get_total_seconds_ts(data)
        best_matching_reference_ts = self._get_matching_reference_ts(data_ts)

        if best_matching_reference_ts is not None:
            self._add_data_by_reference_ts(data, best_matching_reference_ts)
            self._update_ready_timestamps(best_matching_reference_ts)
        else:
            self._unmatched_data.append(data)

    def _get_matching_reference_ts(self, data_ts: float) -> Optional[float]:
        for reference_ts in self._reference_data.keys():
            if self._timestamps_in_tolerance(reference_ts, data_ts):
                return reference_ts
        return None

    def _add_reference(
        self,
        reference: TReference,
    ) -> None:
        reference_ts = self._get_total_seconds_ts(reference)
        self._reference_data[reference_ts] = reference
        self._try_match_data(reference_ts)
        self._update_ready_timestamps(reference_ts)

    def _try_match_data(self, reference_ts: float) -> None:
        matched_data: List[TGathered] = []
        for data in self._unmatched_data:
            data_ts = self._get_total_seconds_ts(data)
            if self._timestamps_in_tolerance(reference_ts, data_ts):
                self._add_data_by_reference_ts(data, reference_ts)
                matched_data.append(data)

        for matched in matched_data:
            self._unmatched_data.remove(matched)

    def _timestamps_in_tolerance(self, timestamp1: float, timestamp2: float) -> bool:
        difference = abs(timestamp1 - timestamp2)
        return difference < (1 / self._camera_fps / self.FPS_TOLERANCE_DIVISOR)

    def _add_data_by_reference_ts(self, data: TGathered, reference_ts: float) -> None:
        if reference_ts in self._data_by_reference_ts:
            self._data_by_reference_ts[reference_ts].append(data)
        else:
            self._data_by_reference_ts[reference_ts] = [data]

    def _update_ready_timestamps(self, timestamp: float) -> None:
        if not self._timestamp_ready(timestamp):
            return

        self._ready_timestamps.put(timestamp)

    def _timestamp_ready(self, timestamp: float) -> bool:
        reference = self._reference_data.get(timestamp)
        if not reference:
            return False

        wait_for_count = self._get_wait_count(reference)
        if wait_for_count == 0:
            return True

        recognitions = self._data_by_reference_ts.get(timestamp)
        if not recognitions:
            return False

        return wait_for_count == len(recognitions)

    def _get_wait_count(self, reference: TReference) -> int:
        return self._wait_count_fn(reference)

    def _pop_ready_data(self) -> Optional[GatheredData]:
        if self._ready_timestamps.empty():
            return None

        timestamp = self._ready_timestamps.get()
        return GatheredData(
            reference_data=self._reference_data.pop(timestamp),
            gathered=self._data_by_reference_ts.pop(timestamp, None) or [],
        )

    def _clear_old_data(self, ready_data: GatheredData) -> None:
        current_timestamp = self._get_total_seconds_ts(ready_data)
        self._clear_unmatched_data(current_timestamp)
        self._clear_old_references(current_timestamp)

    def _clear_unmatched_data(self, current_timestamp: float) -> None:
        unmatched_data_to_remove = []
        for unmatched_data in self._unmatched_data:
            if self._get_total_seconds_ts(unmatched_data) < current_timestamp:
                unmatched_data_to_remove.append(unmatched_data)

        for unmatched_data in unmatched_data_to_remove:
            self._unmatched_data.remove(unmatched_data)

    def _get_total_seconds_ts(self, buffer_like: dai.Buffer) -> float:
        return buffer_like.getTimestamp().total_seconds()

    def _clear_old_references(self, current_timestamp: float) -> None:
        reference_keys_to_pop = []
        for reference_ts in self._reference_data.keys():
            if reference_ts < current_timestamp:
                reference_keys_to_pop.append(reference_ts)

        for reference_ts in reference_keys_to_pop:
            self._reference_data.pop(reference_ts)
            self._data_by_reference_ts.pop(reference_ts, None)

    def set_wait_count_fn(self, fn: Callable[[TReference], int]) -> None:
        self._wait_count_fn = fn



import numpy as np

from depthai_nodes import Map2D


def create_map_message(map: np.ndarray, min_max_scaling: bool = False) -> Map2D:
    """Create a DepthAI message for a map of floats.

    @param map: A NumPy array representing the map with shape HW or NHW/HWN. Here N
        stands for batch dimension.
    @type map: np.array
    @param min_max_scaling: If True, the map is scaled to the range [0, 1]. Defaults to
        False.
    @type min_max_scaling: bool
    @return: An Map2D object containing the density information.
    @rtype: Map2D
    @raise ValueError: If the density map is not a NumPy array.
    @raise ValueError: If the density map is not 2D or 3D.
    @raise ValueError: If the 3D density map shape is not NHW or HWN.
    """

    if not isinstance(map, np.ndarray):
        raise ValueError(f"Expected numpy array, got {type(map)}.")

    if not (len(map.shape) == 2 or len(map.shape) == 3):
        raise ValueError(f"Expected 2D or 3D input, got {len(map.shape)}D input.")

    if len(map.shape) == 3:
        if map.shape[0] == 1:
            map = map[0, :, :]  # NHW to HW
        elif map.shape[2] == 1:
            map = map[:, :, 0]  # HWN to HW
        else:
            raise ValueError(
                f"Unexpected map shape. Expected NHW or HWN, got {map.shape}."
            )

    if min_max_scaling:
        min_val = map.min()
        max_val = map.max()
        if map.min() != map.max():
            map = (map - min_val) / (max_val - min_val)

    if map.dtype != np.float32:
        map = map.astype(np.float32)

    map_2d = Map2D()
    map_2d.map = map

    return map_2d



from typing import List, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.node.parsers.utils import sigmoid


def crop_mask(mask: np.ndarray, bbox: np.ndarray) -> np.ndarray:
    """It takes a mask and a bounding box, and returns a mask that is cropped to the
    bounding box.

    @param mask: [h, w] numpy array of a single mask
    @type mask: np.ndarray
    @param bbox: A numpy array of bbox coordinates in (x_center, y_center, width,
        height) format
    @type bbox: np.ndarray
    @return: A mask that is cropped to the bounding box
    @rtype: np.ndarray
    """
    h, w = mask.shape
    c_x, c_y, width, height = bbox
    x1 = c_x - width / 2
    y1 = c_y - height / 2
    x2 = c_x + width / 2
    y2 = c_y + height / 2
    r = np.arange(w).reshape(1, w)
    c = np.arange(h).reshape(h, 1)

    return mask * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))


def process_single_mask(
    protos: np.ndarray,
    mask_coeff: np.ndarray,
    mask_conf: float,
    bbox: np.ndarray,
) -> np.ndarray:
    """Process a single mask.

    @param protos: Protos.
    @type protos: np.ndarray
    @param mask_coeff: Mask coefficient.
    @type mask_coeff: np.ndarray
    @param mask_conf: Mask confidence.
    @type mask_conf: float
    @param bbox: A numpy array of bbox coordinates in (x_center, y_center, width,
        height) normalized format.
    @type bbox: np.ndarray
    @return: Processed mask.
    @rtype: np.ndarray
    """
    c, mh, mw = protos.shape  # CHW
    scaled_bbox = bbox * np.array([mw, mh, mw, mh])
    mask = sigmoid(np.sum(protos * mask_coeff[..., np.newaxis, np.newaxis], axis=0))
    mask = crop_mask(mask, scaled_bbox)
    return (mask > mask_conf).astype(np.uint8)


def get_segmentation_outputs(
    output: dai.NNData,
    mask_output_layer_names: Optional[List[str]] = None,
    protos_output_layer_name: Optional[str] = None,
) -> Tuple[List[np.ndarray], np.ndarray, int]:
    """Get the segmentation outputs from the Neural Network data."""
    # Get all the layer names
    layer_names = mask_output_layer_names or output.getAllLayerNames()
    mask_outputs = sorted([name for name in layer_names if "mask" in name])
    masks_outputs_values = [
        output.getTensor(
            o, dequantize=True, storageOrder=dai.TensorInfo.StorageOrder.NCHW
        ).astype(np.float32)
        for o in mask_outputs
    ]
    protos_output = output.getTensor(
        protos_output_layer_name or "protos_output",
        dequantize=True,
        storageOrder=dai.TensorInfo.StorageOrder.NCHW,
    ).astype(np.float32)
    protos_len = protos_output.shape[1]
    return masks_outputs_values, protos_output, protos_len



from .input import InfiniteInputMock
from .node import NodeMock
from .queue import OutputQueueMock


class ThreadedHostNodeMock(NodeMock):
    """Mock class for the depthai ThreadedHostNode.

    The class is used to create a mock pipeline for testing purposes.
    """

    def __init__(self):
        super().__init__()
        self._output = OutputQueueMock()
        self._parent_pipeline = None
        self._input = InfiniteInputMock()

    @property
    def input(self):
        return self._input

    @input.setter
    def input(self, value):
        self._input = value

    @property
    def output(self):
        return self._output

    @property
    def out(self):
        return self._output

    @out.setter
    def out(self, node: OutputQueueMock) -> None:
        self._output = node

    @property
    def parentPipeline(self):
        return self._parent_pipeline

    @parentPipeline.setter
    def parentPipeline(self, pipeline):
        self._parent_pipeline = pipeline

    def getParentPipeline(self):
        return self._parent_pipeline



from typing import List, Tuple

import numpy as np

from depthai_nodes.node.parsers.utils import softmax


def decode_ufld(
    anchors: List[int],
    griding_num: int,
    cls_num_per_lane: int,
    input_width: int,
    input_height: int,
    y: np.ndarray,
) -> List[List[Tuple[int, int]]]:
    col_sample = np.linspace(0, input_width - 1, griding_num)
    col_sample_w = col_sample[1] - col_sample[0]

    out_j = y
    out_j = out_j[:, ::-1, :]
    prob = softmax(out_j[:-1, :, :], axis=0)

    idx = np.arange(griding_num) + 1
    idx = idx.reshape(-1, 1, 1)
    loc = np.sum(prob * idx, axis=0)
    out_j = np.argmax(out_j, axis=0)
    loc[out_j == griding_num] = 0
    out_j = loc

    points = [[] for _ in range(out_j.shape[1])]

    for i in range(out_j.shape[1]):
        if np.sum(out_j[:, i] != 0) > 2:
            for k in range(out_j.shape[0]):
                if out_j[k, i] > 0:
                    ppp = [
                        int(out_j[k, i] * col_sample_w) - 1,
                        int(
                            input_height
                            * (anchors[cls_num_per_lane - 1 - k] / input_height)
                        )
                        - 1,
                    ]
                    # Normalize points to [0,1] range
                    ppp[0] /= input_width
                    ppp[1] /= input_height
                    ppp = tuple(ppp)

                    points[i].append(ppp)

    return points



import depthai as dai
import numpy as np
import pytest

from depthai_nodes import SegmentationMask


@pytest.fixture
def segmentation_mask():
    return SegmentationMask()


def test_segmentation_mask_initialization(segmentation_mask: SegmentationMask):
    assert np.array_equal(segmentation_mask.mask, np.array([]))
    assert segmentation_mask.transformation is None


def test_segmentation_mask_set_mask(segmentation_mask: SegmentationMask):
    mask_array = np.random.randint(-1, 256, (480, 640), dtype=np.int16)
    segmentation_mask.mask = mask_array
    assert np.array_equal(segmentation_mask.mask, mask_array)

    with pytest.raises(TypeError):
        segmentation_mask.mask = "not a numpy array"

    with pytest.raises(ValueError):
        segmentation_mask.mask = np.random.randint(
            -1, 256, (480, 640, 3), dtype=np.int16
        )

    with pytest.raises(ValueError):
        segmentation_mask.mask = np.random.randint(-1, 256, (480, 640), dtype=np.uint8)

    with pytest.raises(ValueError):
        segmentation_mask.mask = np.random.randint(-2, 256, (480, 640), dtype=np.int16)


def test_segmentation_mask_set_transformation(segmentation_mask: SegmentationMask):
    transformation = dai.ImgTransformation()
    segmentation_mask.transformation = transformation
    assert segmentation_mask.transformation == transformation

    with pytest.raises(TypeError):
        segmentation_mask.transformation = "not a dai.ImgTransformation"


def test_segmentation_mask_set_transformation_none(segmentation_mask: SegmentationMask):
    segmentation_mask.transformation = None
    assert segmentation_mask.transformation is None



from datetime import timedelta

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.message import GatheredData
from depthai_nodes.node.gather_data import GatherData
from tests.utils import PipelineMock, create_img_detections

DIFFERENT_TESTS = 3  # used for stability tests


@pytest.fixture
def duration(request):
    return request.config.getoption("--duration")


@pytest.fixture
def fps():
    return 30


@pytest.fixture
def tolerance(fps):
    return 1 / fps / GatherData.FPS_TOLERANCE_DIVISOR


@pytest.fixture
def reference_timestamp():
    return timedelta(days=1, hours=1, minutes=1, seconds=1, milliseconds=0)


@pytest.fixture
def in_tolerance_timestamp(reference_timestamp, tolerance):
    return reference_timestamp + timedelta(seconds=(tolerance * 0.9))


@pytest.fixture
def out_of_tolerance_timestamp(reference_timestamp, tolerance):
    return reference_timestamp + timedelta(seconds=(tolerance * 1.1))


@pytest.fixture
def old_reference_timestamp(reference_timestamp, tolerance):
    return reference_timestamp - timedelta(seconds=(tolerance * 1.1))


@pytest.fixture
def gather_data_generator():
    return PipelineMock().create(GatherData)


def create_nn_data(timestamp=None):
    nn_data = dai.NNData()
    tensor = np.random.rand(1, 10).astype(np.float32)
    nn_data.addTensor(name="test_layer", tensor=tensor)
    if timestamp:
        nn_data.setTimestamp(timestamp)
    return nn_data


@pytest.fixture
def nn_data_in_tolerance(in_tolerance_timestamp):
    return create_nn_data(timestamp=in_tolerance_timestamp)


@pytest.fixture
def nn_data_out_of_tolerance(out_of_tolerance_timestamp):
    return create_nn_data(timestamp=out_of_tolerance_timestamp)


@pytest.fixture
def img_detections(reference_timestamp):
    return create_img_detections(timestamp=reference_timestamp)


@pytest.fixture
def old_img_detections(old_reference_timestamp):
    return create_img_detections(timestamp=old_reference_timestamp)


def setup_gather_data_node(generator, fps, duration=None):
    gather_data = generator.build(camera_fps=fps)
    if duration is not None:
        modified_duration = duration / DIFFERENT_TESTS
        gather_data.input_data._queue.duration = modified_duration
        gather_data.input_reference._queue.duration = modified_duration
    return gather_data


def send_data_and_run(gather_data, reference_data, data_items, checking_function=None):
    gather_data.out.createOutputQueue(checking_function=checking_function)
    gather_data.input_reference.send(reference_data)
    for item in data_items:
        gather_data.input_data.send(item)
    gather_data.run()


def assert_gather_data_result_basics(
    result, expected_reference, expected_gathered_count
):
    assert isinstance(result, GatheredData)
    assert result.reference_data == expected_reference
    assert len(result.gathered) == expected_gathered_count
    assert result.getTimestamp() == expected_reference.getTimestamp()


def assert_timestamps_in_tolerance(timestamps, reference_timestamp, tolerance):
    tolerance_td = timedelta(seconds=tolerance)
    for timestamp in timestamps:
        assert (
            reference_timestamp - tolerance_td
            <= timestamp
            <= reference_timestamp + tolerance_td
        )


def assert_results_not_empty(results):
    assert isinstance(results, list)
    assert len(results) > 0


def assert_gather_data_result(
    result, expected_reference, expected_gathered_count, reference_timestamp, tolerance
):
    assert isinstance(result, GatheredData)
    assert result.reference_data == expected_reference
    assert len(result.gathered) == expected_gathered_count
    assert result.getTimestamp() == expected_reference.getTimestamp()
    gathered_timestamps = [gathered.getTimestamp() for gathered in result.gathered]
    tolerance_td = timedelta(seconds=tolerance)
    for timestamp in gathered_timestamps:
        assert (
            reference_timestamp - tolerance_td
            <= timestamp
            <= reference_timestamp + tolerance_td
        )
    for gathered in result.gathered:
        assert isinstance(gathered, dai.Buffer)


def test_build(gather_data_generator, fps):
    with pytest.raises(ValueError):
        gather_data_generator.build(camera_fps=-fps)

    gather_data = gather_data_generator.build(camera_fps=fps)
    assert gather_data is not None


def test_run_without_build(gather_data_generator):
    with pytest.raises(ValueError):
        gather_data_generator.run()


def test_img_detections(
    gather_data_generator,
    fps,
    img_detections,
    nn_data_in_tolerance,
    nn_data_out_of_tolerance,
    reference_timestamp,
    tolerance,
    duration,
):
    def check(result, *_):
        assert_gather_data_result(
            result=result,
            expected_reference=img_detections,
            expected_gathered_count=3,
            reference_timestamp=reference_timestamp,
            tolerance=tolerance,
        )

    gather_data = setup_gather_data_node(
        generator=gather_data_generator, fps=fps, duration=duration
    )
    data_items = [nn_data_in_tolerance, nn_data_in_tolerance, nn_data_out_of_tolerance]
    send_data_and_run(
        gather_data=gather_data,
        reference_data=img_detections,
        data_items=data_items,
        checking_function=check,
    )


def test_set_wait_count_fn(gather_data_generator, fps, duration, nn_data_in_tolerance):
    def check(result, *_):
        assert isinstance(result, GatheredData)
        assert result.reference_data == nn_data_in_tolerance
        assert len(result.gathered) == 1
        assert result.gathered[0] == nn_data_in_tolerance

    gather_data = setup_gather_data_node(
        generator=gather_data_generator, fps=fps, duration=duration
    )
    gather_data.set_wait_count_fn(lambda _: 1)
    send_data_and_run(
        gather_data=gather_data,
        reference_data=nn_data_in_tolerance,
        data_items=[nn_data_in_tolerance],
        checking_function=check,
    )


def test_clear_old_data(
    gather_data_generator,
    fps,
    duration,
    old_img_detections,
    img_detections,
    nn_data_in_tolerance,
):
    def check(result, *_):
        assert isinstance(result, GatheredData)
        assert result.reference_data == img_detections
        assert len(result.gathered) == 3

    gather_data = setup_gather_data_node(
        generator=gather_data_generator, fps=fps, duration=duration
    )
    gather_data.input_reference.send(old_img_detections)
    gather_data.input_reference.send(img_detections)
    data_items = [nn_data_in_tolerance, nn_data_in_tolerance]
    send_data_and_run(
        gather_data=gather_data,
        reference_data=img_detections,
        data_items=data_items,
        checking_function=check,
    )



SLUGS = [
    "yunet:new-240x320",
    "mediapipe-hand-landmarker:224x224",
    "age-gender-recognition:new-62x62",
    "paddle-text-recognition:160x48",
    "midas-v2-1:small-192x256",
    "paddle-text-detection:256x256",
    "deeplab-v3-plus:256x256",
    "yolov8-nano-pose-estimation:coco-512x288",
    "yolov8-instance-segmentation-nano:coco-512x288",
    "ewasr:384x512",
    "m-lsd:512x512",
]
# SLUGS = [
#     "yunet:new-new-480x640",
#     "osnet:market1501-128x256",
#     "osnet:imagenet-128x256",
#     "osnet:multi-source-domain-128x256",
#     "arcface:lfw-112x112",
#     "mediapipe-hand-landmarker:224x224",
#     "mediapipe-hand-landmarker:old-224x224",
#     "head-pose-estimation:60x60",
#     "yunet:new-960x1280",
#     "age-gender-recognition:new-62x62",
#     "yunet:new-240x320",
#     "yunet:new-480x640",
#     "yunet:new-720x960",
#     "yunet:320x320",
#     "yunet:640x640",
#     "lite-hrnet:30-coco-384x288",
#     "lite-hrnet:30-coco-256x192",
#     "lite-hrnet:18-coco-384x288",
#     "lite-hrnet:18-coco-256x192",
#     "xfeat:stereo-480x640",
#     "xfeat:mono-480x640",
#     "xfeat:stereo-240x320",
#     "xfeat:mono-240x320",
#     "objectron:camera-224x224",
#     "objectron:chair-224x224",
#     "objectron:cup-224x224",
#     "objectron:sneakers-224x224",
#     "xfeat:240x320",
#     "xfeat:480x640",
#     "depth-anything-v2:vit-s-mde-outdoors-252x336",
#     "depth-anything-v2:vit-s-mde-indoors-252x336",
#     "depth-anything-v2:vit-s-420x560",
#     "depth-anything-v2:vit-s-252x336",
#     "paddle-text-recognition:160x48",
#     "midas-v2-1:small-768x1024",
#     "midas-v2-1:small-384x512",
#     "midas-v2-1:small-192x256",
#     "dm-count:sha-144x256",
#     "dm-count:sha-240x426",
#     "dm-count:sha-360x640",
#     "dm-count:sha-540x960",
#     "dm-count:shb-144x256",
#     "dm-count:shb-240x426",
#     "dm-count:shb-360x640",
#     "dm-count:shb-540x960",
#     "dm-count:qnrf-144x256",
#     "dm-count:qnrf-240x426",
#     "dm-count:qnrf-360x640",
#     "dm-count:qnrf-540x960",
#     "ppe-detection:640x640",
#     "midas-v2-1:small-288x512",
#     "midas-v2-1:small-256x384",
#     "paddle-text-recognition:320x48",
#     "paddle-text-detection:256x256",
#     "paddle-text-detection:544x960",
#     "paddle-text-detection:320x576",
#     "ultra-fast-lane-detection:culane-800x288",
#     "ultra-fast-lane-detection:tusimple-800x288",
#     "deeplab-v3-plus:256x256",
#     "deeplab-v3-plus:person-513x513",
#     "deeplab-v3-plus:person-256x256",
#     "deeplab-v3-plus:513x513",
#     "yolov6-nano:coco-416x416",
#     "pp-liteseg:512x1024",
#     "yolov8-large-pose-estimation:coco-640x352",
#     "yolov8-nano-pose-estimation:coco-512x288",
#     "yolov8-instance-segmentation-nano:coco-512x288",
#     "yolov8-instance-segmentation-large:coco-640x352",
#     "fastsam-x:640x352",
#     "fastsam-s:512x288",
#     "yolov10-nano:coco-512x288",
#     "ewasr:384x512",
#     "m-lsd:512x512",
#     "m-lsd-tiny:512x512",
#     "fastsam-s:clip-visual",
#     "scrfd-person-detection:2-5g-640x640",
#     "emotion-recognition:gray-64x64",
#     "efficientnet-lite:lite4-300x300",
#     "efficientnet-lite:lite0-224x224",
#     "mobilenet-ssd:300x300",
#     "license-plate-detection:640x640",
#     "vehicle-attributes-classification:72x72",
#     "emotion-recognition:260x260",
#     "age-gender-recognition:62x62",
#     "fastsam-s:clip-textual",
#     "mediapipe-palm-detection:128x128",
#     "xfeat:352x640",
#     "esrgan:256x256",
#     "rt-super-resolution:50x50",
#     "qrdet:nano-288x512",
#     "yolov6-large:r2-coco-640x352",
#     "yolov6-nano:r2-adjusted-coco-512x288",
#     "yolov6-nano:r2-coco-512x288",
#     "zero-dce:400x600",
#     "dncnn3:321x481",
#     "mediapipe-selfie-segmentation:144x256",
#     "superanimal-landmarker:256x256",
#     "mediapipe-palm-detection:192x192",
#     "mediapipe-face-landmarker:192x192",
#     "scrfd-face-detection:10g-640x640",
# ]

PARSERS_SLUGS = {
    "YuNetParser": [
        "yunet:new-new-480x640",
        "yunet:new-960x1280",
        "yunet:new-240x320",
        "yunet:new-480x640",
        "yunet:new-720x960",
        "yunet:320x320",
        "yunet:640x640",
    ],
    "EmbeddingsParser": [
        "osnet:market1501-128x256",
        "osnet:imagenet-128x256",
        "osnet:multi-source-domain-128x256",
        "arcface:lfw-112x112",
    ],
    "KeypointParser": [
        "mediapipe-hand-landmarker:224x224",
        "objectron:camera-224x224",
        "objectron:chair-224x224",
        "objectron:cup-224x224",
        "objectron:sneakers-224x224",
        "mediapipe-face-landmarker:192x192",
    ],
    "RegressionParser": [
        "mediapipe-hand-landmarker:224x224",
        "mediapipe-hand-landmarker:224x224",
        "head-pose-estimation:60x60",
        "head-pose-estimation:60x60",
        "head-pose-estimation:60x60",
        "age-gender-recognition:new-62x62",
        "objectron:camera-224x224",
        "objectron:chair-224x224",
        "objectron:cup-224x224",
        "objectron:sneakers-224x224",
    ],
    "MPHandLandmarkParser": ["mediapipe-hand-landmarker:old-224x224"],
    "ClassificationParser": [
        "age-gender-recognition:new-62x62",
        "emotion-recognition:gray-64x64",
        "efficientnet-lite:lite4-300x300",
        "efficientnet-lite:lite0-224x224",
        "vehicle-attributes-classification:72x72",
        "vehicle-attributes-classification:72x72",
        "emotion-recognition:260x260",
    ],
    "HRNetParser": [
        "lite-hrnet:30-coco-384x288",
        "lite-hrnet:30-coco-256x192",
        "lite-hrnet:18-coco-384x288",
        "lite-hrnet:18-coco-256x192",
    ],
    "XFeatStereoParser": ["xfeat:stereo-480x640", "xfeat:stereo-240x320"],
    "XFeatMonoParser": ["xfeat:mono-480x640", "xfeat:mono-240x320"],
    "XFeatParser": ["xfeat:240x320", "xfeat:480x640", "xfeat:352x640"],
    "MapOutputParser": [
        "depth-anything-v2:vit-s-mde-outdoors-252x336",
        "depth-anything-v2:vit-s-mde-indoors-252x336",
        "depth-anything-v2:vit-s-420x560",
        "depth-anything-v2:vit-s-252x336",
        "midas-v2-1:small-768x1024",
        "midas-v2-1:small-384x512",
        "midas-v2-1:small-192x256",
        "dm-count:sha-144x256",
        "dm-count:sha-240x426",
        "dm-count:sha-360x640",
        "dm-count:sha-540x960",
        "dm-count:shb-144x256",
        "dm-count:shb-240x426",
        "dm-count:shb-360x640",
        "dm-count:shb-540x960",
        "dm-count:qnrf-144x256",
        "dm-count:qnrf-240x426",
        "dm-count:qnrf-360x640",
        "dm-count:qnrf-540x960",
        "midas-v2-1:small-288x512",
        "midas-v2-1:small-256x384",
    ],
    "ClassificationSequenceParser": [
        "paddle-text-recognition:160x48",
        "paddle-text-recognition:320x48",
    ],
    "YOLO": [
        "ppe-detection:640x640",
        "yolov6-nano:coco-416x416",
        "yolov10-nano:coco-512x288",
        "license-plate-detection:640x640",
        "qrdet:nano-288x512",
        "yolov6-large:r2-coco-640x352",
        "yolov6-nano:r2-adjusted-coco-512x288",
        "yolov6-nano:r2-coco-512x288",
    ],
    "PPTextDetectionParser": [
        "paddle-text-detection:256x256",
        "paddle-text-detection:544x960",
        "paddle-text-detection:320x576",
    ],
    "LaneDetectionParser": [
        "ultra-fast-lane-detection:culane-800x288",
        "ultra-fast-lane-detection:tusimple-800x288",
    ],
    "SegmentationParser": [
        "deeplab-v3-plus:256x256",
        "deeplab-v3-plus:person-513x513",
        "deeplab-v3-plus:person-256x256",
        "deeplab-v3-plus:513x513",
        "pp-liteseg:512x1024",
        "ewasr:384x512",
        "mediapipe-selfie-segmentation:144x256",
    ],
    "YOLOExtendedParser": [
        "yolov8-large-pose-estimation:coco-640x352",
        "yolov8-nano-pose-estimation:coco-512x288",
        "yolov8-instance-segmentation-nano:coco-512x288",
        "yolov8-instance-segmentation-large:coco-640x352",
    ],
    "FastSAMParser": ["fastsam-x:640x352", "fastsam-s:512x288"],
    "MLSDParser": ["m-lsd:512x512", "m-lsd-tiny:512x512"],
    "SCRFDParser": [
        "scrfd-person-detection:2-5g-640x640",
        "scrfd-face-detection:10g-640x640",
    ],
    "SSD": ["mobilenet-ssd:300x300"],
    "AgeGenderParser": ["age-gender-recognition:62x62"],
    "MPPalmDetectionParser": [
        "mediapipe-palm-detection:128x128",
        "mediapipe-palm-detection:192x192",
    ],
    "ImageOutputParser": [
        "esrgan:256x256",
        "rt-super-resolution:50x50",
        "zero-dce:400x600",
        "dncnn3:321x481",
    ],
    "SuperAnimalParser": ["superanimal-landmarker:256x256"],
}



import copy
from typing import List

import cv2
import depthai as dai
import numpy as np

from .utils import (
    copy_message,
)


class Cluster(dai.Buffer):
    """Cluster class for storing a cluster.

    Attributes
    ----------
    label : int
        Label of the cluster.
    points : List[dai.Point2f]
        List of points in the cluster.
    """

    def __init__(self):
        """Initializes the Cluster object."""
        super().__init__()
        self._label: int = None
        self._points: List[dai.Point2f] = []

    def copy(self):
        """Creates a new instance of the Cluster class and copies the attributes.

        @return: A new instance of the Cluster class.
        @rtype: Cluster
        """
        new_obj = Cluster()
        new_obj.label = copy.deepcopy(self.label)
        new_obj.points = [copy_message(p) for p in self.points]
        return new_obj

    @property
    def label(self) -> int:
        """Returns the label of the cluster.

        @return: Label of the cluster.
        @rtype: int
        """
        return self._label

    @label.setter
    def label(self, value: int):
        """Sets the label of the cluster.

        @param value: Label of the cluster.
        @type value: int
        @raise TypeError: If value is not an int.
        """
        if not isinstance(value, int):
            raise TypeError(f"Label must be of type int, instead got {type(value)}.")
        self._label = value

    @property
    def points(self) -> List[dai.Point2f]:
        """Returns the points in the cluster.

        @return: List of points in the cluster.
        @rtype: List[dai.Point2f]
        """
        return self._points

    @points.setter
    def points(self, value: List[dai.Point2f]):
        """Sets the points in the cluster.

        @param value: List of points in the cluster.
        @type value: List[dai.Point2f]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each element is not of type dai.Point2f.
        """
        if not isinstance(value, List):
            raise TypeError(f"Points must be a list, instead got {type(value)}.")
        if not all(isinstance(point, dai.Point2f) for point in value):
            raise ValueError("Points must be a list of dai.Point2f objects")
        self._points = value


class Clusters(dai.Buffer):
    """Clusters class for storing clusters.

    Attributes
    ----------
    clusters : List[Cluster]
        List of clusters.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self):
        """Initializes the Clusters object."""
        super().__init__()
        self._clusters: List[Cluster] = []
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the Clusters class and copies the attributes.

        @return: A new instance of the Clusters class.
        @rtype: Clusters
        """
        new_obj = Clusters()
        new_obj.clusters = [cluster.copy() for cluster in self.clusters]
        new_obj.setSequenceNum(self.getSequenceNum())
        new_obj.setTimestamp(self.getTimestamp())
        new_obj.setTimestampDevice(self.getTimestampDevice())
        new_obj.setTransformation(self.transformation)
        return new_obj

    @property
    def clusters(self) -> List[Cluster]:
        """Returns the clusters.

        @return: List of clusters.
        @rtype: List[Cluster]
        """
        return self._clusters

    @clusters.setter
    def clusters(self, value: List[Cluster]):
        """Sets the clusters.

        @param value: List of clusters.
        @type value: List[Cluster]
        @raise TypeError: If value is not a list.
        @raise ValueError: If each element is not of type Cluster.
        """
        if not isinstance(value, List):
            raise TypeError("Clusters must be a list.")
        if not all(isinstance(cluster, Cluster) for cluster in value):
            raise ValueError("Clusters must be a list of Cluster objects.")
        self._clusters = value

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """

        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )
        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        self.transformation = transformation

    def getVisualizationMessage(self) -> dai.ImgAnnotations:
        """Creates a default visualization message for clusters and colors each one
        separately."""
        img_annotations = dai.ImgAnnotations()
        annotation = dai.ImgAnnotation()

        num_clusters = len(self.clusters)
        color_mask = np.array(range(0, 255, 255 // num_clusters), dtype=np.uint8)
        color_mask = cv2.applyColorMap(color_mask, cv2.COLORMAP_RAINBOW)
        color_mask = color_mask / 255
        color_mask = color_mask.reshape(-1, 3)

        for i, cluster in enumerate(self.clusters):
            pointsAnnotation = dai.PointsAnnotation()
            pointsAnnotation.type = dai.PointsAnnotationType.POINTS
            pointsAnnotation.points = dai.VectorPoint2f(cluster.points)
            r, g, b = color_mask[i]
            color = dai.Color(r, g, b)
            pointsAnnotation.outlineColor = color
            pointsAnnotation.fillColor = color
            pointsAnnotation.thickness = 2.0
            annotation.points.append(pointsAnnotation)

        img_annotations.annotations.append(annotation)
        img_annotations.setTimestamp(self.getTimestamp())
        return img_annotations



from typing import Any, Dict, List, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_detection_message
from depthai_nodes.node.parsers.detection import DetectionParser
from depthai_nodes.node.parsers.utils import xyxy_to_xywh
from depthai_nodes.node.parsers.utils.scrfd import decode_scrfd


class SCRFDParser(DetectionParser):
    """Parser class for parsing the output of the SCRFD face detection model.

    Attributes
    ----------
    output_layer_name: List[str]
        Names of the output layers relevant to the parser.
    conf_threshold : float
        Confidence score threshold for detected faces.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.
    input_size : tuple
        Input size of the model.
    feat_stride_fpn : tuple
        Tuple of the feature strides.
    num_anchors : int
        Number of anchors.

    Output Message/s
    ----------------
    **Type**: dai.ImgDetections

    **Description**: ImgDetections message containing bounding boxes, labels, and confidence scores of detected faces.
    """

    def __init__(
        self,
        output_layer_names: List[str] = None,
        conf_threshold: float = 0.5,
        iou_threshold: float = 0.5,
        max_det: int = 100,
        input_size: Tuple[int, int] = (640, 640),
        feat_stride_fpn: Tuple = (8, 16, 32),
        num_anchors: int = 2,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_names: Names of the output layers relevant to the parser.
        @type output_layer_names: List[str]
        @param conf_threshold: Confidence score threshold for detected faces.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        @param input_size: Input size of the model.
        @type input_size: tuple
        @param feat_stride_fpn: List of the feature strides.
        @type feat_stride_fpn: tuple
        @param num_anchors: Number of anchors.
        @type num_anchors: int
        """
        super().__init__(conf_threshold, iou_threshold, max_det)
        self.output_layer_names = (
            [] if output_layer_names is None else output_layer_names
        )

        self.feat_stride_fpn = feat_stride_fpn
        self.num_anchors = num_anchors
        self.input_size = input_size
        self.label_names = ["Face"]
        self._logger.debug(
            f"SCRFDParser initialized with output_layer_names={output_layer_names}, conf_threshold={conf_threshold}, iou_threshold={iou_threshold}, max_det={max_det}, input_size={input_size}, feat_stride_fpn={feat_stride_fpn}, num_anchors={num_anchors}"
        )

    def setOutputLayerNames(self, output_layer_names: List[str]) -> None:
        """Sets the output layer name(s) for the parser.

        @param output_layer_names: The name of the output layer(s) to be used.
        @type output_layer_names: List[str]
        """
        if not isinstance(output_layer_names, list):
            raise ValueError("Output layer names must be a list.")
        if not all(isinstance(layer, str) for layer in output_layer_names):
            raise ValueError("Output layer names must be a list of strings.")
        self.output_layer_names = output_layer_names
        self._logger.debug(f"Output layer names set to {self.output_layer_names}")

    def setInputSize(self, input_size: Tuple[int, int]) -> None:
        """Sets the input size of the model.

        @param input_size: Input size of the model.
        @type input_size: list
        """
        if not isinstance(input_size, tuple):
            raise ValueError("Input size must be a tuple.")
        if not all(isinstance(size, int) for size in input_size):
            raise ValueError("Input size must be a tuple of integers.")
        self.input_size = input_size
        self._logger.debug(f"Input size set to {self.input_size}")

    def setFeatStrideFPN(self, feat_stride_fpn: List[int]) -> None:
        """Sets the feature stride of the FPN.

        @param feat_stride_fpn: Feature stride of the FPN.
        @type feat_stride_fpn: list
        """
        if not isinstance(feat_stride_fpn, list):
            raise ValueError("Feature stride must be a list.")
        if not all(isinstance(stride, int) for stride in feat_stride_fpn):
            raise ValueError("Feature stride must be a list of integers.")
        self.feat_stride_fpn = feat_stride_fpn
        self._logger.debug(f"Feature stride set to {self.feat_stride_fpn}")

    def setNumAnchors(self, num_anchors: int) -> None:
        """Sets the number of anchors.

        @param num_anchors: Number of anchors.
        @type num_anchors: int
        """
        if not isinstance(num_anchors, int):
            raise ValueError("Number of anchors must be an integer.")
        self.num_anchors = num_anchors
        self._logger.debug(f"Number of anchors set to {self.num_anchors}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "SCRFDParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: SCRFDParser
        """

        super().build(head_config)

        output_layers = head_config.get("outputs", [])
        score_layer_names = [layer for layer in output_layers if "score" in layer]
        bbox_layer_names = [layer for layer in output_layers if "bbox" in layer]
        kps_layer_names = [layer for layer in output_layers if "kps" in layer]

        if len(score_layer_names) != len(bbox_layer_names) or len(
            score_layer_names
        ) != len(kps_layer_names):
            raise ValueError(
                f"Number of score, bbox, and kps layers should be equal, got {len(score_layer_names)}, {len(bbox_layer_names)}, and {len(kps_layer_names)} layers."
            )

        self.output_layer_names = output_layers
        self.feat_stride_fpn = head_config.get("feat_stride_fpn", self.feat_stride_fpn)
        self.num_anchors = head_config.get("num_anchors", self.num_anchors)

        self._logger.debug(
            f"SCRFDParser built with output_layer_names={self.output_layer_names}, feat_stride_fpn={self.feat_stride_fpn}, num_anchors={self.num_anchors}"
        )

        return self

    def run(self):
        self._logger.debug("SCRFDParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            scores_concatenated = []
            bboxes_concatenated = []
            kps_concatenated = []

            if len(self.output_layer_names) == 0:
                self.output_layer_names = output.getAllLayerNames()

            self._logger.debug(
                f"Processing input with layers: {output.getAllLayerNames()}"
            )

            for stride in self.feat_stride_fpn:
                score_layer_name = f"score_{stride}"
                bbox_layer_name = f"bbox_{stride}"
                kps_layer_name = f"kps_{stride}"
                if score_layer_name not in self.output_layer_names:
                    raise ValueError(
                        f"Layer {score_layer_name} not found in the model output."
                    )
                if bbox_layer_name not in self.output_layer_names:
                    raise ValueError(
                        f"Layer {bbox_layer_name} not found in the model output."
                    )
                if kps_layer_name not in self.output_layer_names:
                    raise ValueError(
                        f"Layer {kps_layer_name} not found in the model output."
                    )

                score_tensor = (
                    output.getTensor(score_layer_name, dequantize=True)
                    .flatten()
                    .astype(np.float32)
                )
                bbox_tensor = (
                    output.getTensor(bbox_layer_name, dequantize=True)
                    .reshape(len(score_tensor), 4)
                    .astype(np.float32)
                )
                kps_tensor = (
                    output.getTensor(kps_layer_name, dequantize=True)
                    .reshape(len(score_tensor), 10)
                    .astype(np.float32)
                )

                scores_concatenated.append(score_tensor)
                bboxes_concatenated.append(bbox_tensor)
                kps_concatenated.append(kps_tensor)

            bboxes, scores, keypoints = decode_scrfd(
                bboxes_concatenated=bboxes_concatenated,
                scores_concatenated=scores_concatenated,
                kps_concatenated=kps_concatenated,
                feat_stride_fpn=self.feat_stride_fpn,
                input_size=self.input_size,
                num_anchors=self.num_anchors,
                score_threshold=self.conf_threshold,
                nms_threshold=self.iou_threshold,
            )
            bboxes = xyxy_to_xywh(bboxes)
            bboxes = np.clip(bboxes, 0, 1)

            labels = np.array([0] * len(bboxes))

            label_names = (
                [self.label_names[label] for label in labels]
                if self.label_names
                else None
            )
            message = create_detection_message(
                bboxes=bboxes,
                scores=scores,
                labels=labels,
                label_names=label_names,
                keypoints=keypoints,
            )
            message.setTimestamp(output.getTimestamp())
            message.setTransformation(output.getTransformation())
            message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created detection message with {len(bboxes)} detections"
            )

            self.out.send(message)

            self._logger.debug("Detection message sent successfully")



import depthai as dai
import numpy as np
import pytest

from depthai_nodes.node.host_spatials_calc import HostSpatialsCalc

from .utils.calibration_handler import get_calibration_handler


@pytest.fixture
def calib_data():
    return get_calibration_handler()


@pytest.fixture
def depth_frame():
    frame = np.random.randint(200, 30000, (720, 1280), dtype=np.uint16)
    img_frame = dai.ImgFrame()
    img_frame.setCvFrame(frame, dai.ImgFrame.Type.RAW16)
    img_frame.setWidth(1280)
    img_frame.setHeight(720)
    return img_frame


@pytest.fixture
def host_spatials_calc(calib_data):
    return HostSpatialsCalc(calib_data)


def test_set_lower_threshold(host_spatials_calc: HostSpatialsCalc):
    host_spatials_calc.setLowerThreshold(100)
    assert host_spatials_calc.thresh_low == 100

    host_spatials_calc.setLowerThreshold(150.5)
    assert host_spatials_calc.thresh_low == 150

    with pytest.raises(TypeError):
        host_spatials_calc.setLowerThreshold("invalid")


def test_set_upper_threshold(host_spatials_calc: HostSpatialsCalc):
    host_spatials_calc.setUpperThreshold(50000)
    assert host_spatials_calc.thresh_high == 50000

    host_spatials_calc.setUpperThreshold(25000.5)
    assert host_spatials_calc.thresh_high == 25000

    with pytest.raises(TypeError):
        host_spatials_calc.setUpperThreshold("invalid")


def test_set_delta_roi(host_spatials_calc: HostSpatialsCalc):
    host_spatials_calc.setDeltaRoi(10)
    assert host_spatials_calc.delta == 10

    host_spatials_calc.setDeltaRoi(15.5)
    assert host_spatials_calc.delta == 15

    with pytest.raises(TypeError):
        host_spatials_calc.setDeltaRoi("invalid")


def test_check_input_roi(
    host_spatials_calc: HostSpatialsCalc, depth_frame: dai.ImgFrame
):
    roi = [100, 100, 200, 200]
    result = host_spatials_calc._check_input(roi, depth_frame.getFrame())
    assert result == roi


def test_check_input_point(
    host_spatials_calc: HostSpatialsCalc, depth_frame: dai.ImgFrame
):
    point = [100, 100]
    result = host_spatials_calc._check_input(point, depth_frame.getFrame())
    assert result == [95, 95, 105, 105]


def test_check_input_invalid(
    host_spatials_calc: HostSpatialsCalc, depth_frame: dai.ImgFrame
):
    with pytest.raises(ValueError):
        host_spatials_calc._check_input([100], depth_frame.getFrame())


def test_calc_spatials(host_spatials_calc: HostSpatialsCalc, depth_frame: dai.ImgFrame):
    roi = [100, 100, 200, 200]
    spatials = host_spatials_calc.calc_spatials(depth_frame, roi)
    assert "x" in spatials
    assert "y" in spatials
    assert "z" in spatials
    assert isinstance(spatials["x"], float)
    assert isinstance(spatials["y"], float)
    assert isinstance(spatials["z"], float)



from typing import List, Optional


def generate_script_content(
    resize_width: int,
    resize_height: int,
    resize_mode: str = "STRETCH",
    padding: float = 0,
    valid_labels: Optional[List[int]] = None,
) -> str:
    """The function generates the script content for the dai.Script node.

    It is used to crop and resize the input image based on the detected object. It can
    also work with padding around the detection bounding box and filter detections by
    labels.

    @param resize_width: Target width for the resized image
    @type resize_width: int
    @param resize_height: Target height for the resized image
    @type resize_height: int
    @param resize_mode: Resize mode for the image. Supported values: "CENTER_CROP",
        "LETTERBOX", "NONE", "STRETCH". Default: "STRETCH".
        "STRETCH" - stretches the image so that the corners of the region are now in the corners of the output image.
        "CENTER_CROP" - resizes + crops the image to keep aspect ratio and fill the final size.
        "LETTERBOX" - resizes + pads the image to the final size to keep aspect ratio.
        "NONE" - does not scale and pads top, bottom, left and right to fill the final image.
    @type resize_mode: str
    @param padding: Additional padding around the detection in normalized coordinates
        (0-1)
    @type padding: float
    @param valid_labels: List of valid label indices to filter detections. If None, all
        detections are processed
    @type valid_labels: Optional[List[int]]
    @return: Generated script content as a string
    @rtype: str
    """

    if resize_mode not in ["CENTER_CROP", "LETTERBOX", "NONE", "STRETCH"]:
        raise ValueError("Unsupported resize mode")

    cfg_content = f"""
            cfg = ImageManipConfig()
            rect = RotatedRect()
            rect.center.x = (det.xmin + det.xmax) / 2
            rect.center.y = (det.ymin + det.ymax) / 2
            rect.size.width = det.xmax - det.xmin
            rect.size.height = det.ymax - det.ymin
            rect.size.width = rect.size.width + {padding} * 2
            rect.size.height = rect.size.height + {padding} * 2
            rect.angle = 0

            cfg.addCropRotatedRect(rect, True)
            cfg.setOutputSize({resize_width}, {resize_height}, ImageManipConfig.ResizeMode.{resize_mode})
        """
    validate_label = (
        f"""
            if det.label not in {valid_labels}:
                continue
        """
        if valid_labels
        else ""
    )
    return f"""
try:
    while True:
        frame = node.inputs['preview'].get()
        dets = node.inputs['det_in'].get()

        for i, det in enumerate(dets.detections):
            {validate_label.strip()}

            {cfg_content.strip()}

            node.outputs['manip_cfg'].send(cfg)
            node.outputs['manip_img'].send(frame)

except Exception as e:
    node.warn(str(e))
"""



def pytest_addoption(parser):
    parser.addoption("--nn_archive_path", default="", type=str, help="NN archive path")
    parser.addoption("--model", default="", type=str, help="Model from the ZOO.")
    parser.addoption(
        "--platform",
        default="",
        type=str,
        help="RVC platform to run the tests on.",
    )
    parser.addoption("--models", default="", type=str, help="Model slug from the ZOO.")
    parser.addoption("--parsers", default="", type=str, help="Parsers to test.")
    parser.addoption(
        "--duration",
        default=None,
        type=int,
        help="Duration of the test in seconds.",
    )



from typing import Any, Dict

import depthai as dai

from depthai_nodes.message.creators import create_map_message
from depthai_nodes.node.parsers.base_parser import BaseParser


class MapOutputParser(BaseParser):
    """A parser class for models that produce map outputs, such as depth maps (e.g.
    DepthAnything), density maps (e.g. DM-Count), heat maps, and similar.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    min_max_scaling : bool
        If True, the map is scaled to the range [0, 1].

    Output Message/s
    ----------------
    **Type**: Map2D

    **Description**: Density message containing the density map. The density map is represented with Map2D object.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        min_max_scaling: bool = False,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param min_max_scaling: If True, the map is scaled to the range [0, 1].
        @type min_max_scaling: bool
        """
        super().__init__()
        self.min_max_scaling = min_max_scaling
        self.output_layer_name = output_layer_name
        self._logger.debug(
            f"MapOutputParser initialized with output_layer_name='{output_layer_name}', min_max_scaling={min_max_scaling}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setMinMaxScaling(self, min_max_scaling: bool) -> None:
        """Sets the min_max_scaling flag.

        @param min_max_scaling: If True, the map is scaled to the range [0, 1].
        @type min_max_scaling: bool
        """
        if not isinstance(min_max_scaling, bool):
            raise ValueError("min_max_scaling must be a boolean.")
        self.min_max_scaling = min_max_scaling
        self._logger.debug(f"Min max scaling set to {self.min_max_scaling}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "MapOutputParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: MapOutputParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"MapOutputParser expects exactly 1 output layers, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]
        self.min_max_scaling = head_config.get("min_max_scaling", self.min_max_scaling)

        self._logger.debug(
            f"MapOutputParser built with output_layer_name='{self.output_layer_name}', min_max_scaling={self.min_max_scaling}"
        )

        return self

    def run(self):
        self._logger.debug("MapOutputParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            map = output.getTensor(self.output_layer_name, dequantize=True)

            if map.shape[0] == 1:
                map = map[0]  # remove batch dimension

            map_message = create_map_message(
                map=map, min_max_scaling=self.min_max_scaling
            )
            map_message.setTimestamp(output.getTimestamp())
            map_message.setTransformation(output.getTransformation())
            map_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug("Created Map message.")

            self.out.send(map_message)

            self._logger.debug("Map message sent successfully")
